import asyncio
import aioredis


async def main():
    sentinel_client = await aioredis.create_sentinel(
        [('localhost', 26379)])

    master_redis = sentinel_client.master_for('mymaster')
    info = await master_redis.role()
    print("Master role:", info)
    assert info.role == 'master'

    sentinel_client.close()
    await sentinel_client.wait_closed()


if __name__ == '__main__':
    asyncio.get_event_loop().run_until_complete(main())



import asyncio
import pytest

from aioredis import ReplyError, MultiExecError, WatchVariableError


@pytest.mark.run_loop
def test_multi_exec(redis, loop):
    yield from redis.delete('foo', 'bar')

    tr = redis.multi_exec()
    f1 = tr.incr('foo')
    f2 = tr.incr('bar')
    res = yield from tr.execute()
    assert res == [1, 1]
    res2 = yield from asyncio.gather(f1, f2, loop=loop)
    assert res == res2

    tr = redis.multi_exec()
    f1 = tr.incr('foo')
    f2 = tr.incr('bar')
    yield from tr.execute()
    assert (yield from f1) == 2
    assert (yield from f2) == 2

    tr = redis.multi_exec()
    f1 = tr.set('foo', 1.0)
    f2 = tr.incrbyfloat('foo', 1.2)
    res = yield from tr.execute()
    assert res == [True, 2.2]
    res2 = yield from asyncio.gather(f1, f2, loop=loop)
    assert res == res2

    tr = redis.multi_exec()
    f1 = tr.incrby('foo', 1.0)
    with pytest.raises(MultiExecError, match="increment must be .* int"):
        yield from tr.execute()
    with pytest.raises(TypeError):
        yield from f1


@pytest.mark.run_loop
def test_empty(redis):
    tr = redis.multi_exec()
    res = yield from tr.execute()
    assert res == []


@pytest.mark.run_loop
def test_double_execute(redis):
    tr = redis.multi_exec()
    yield from tr.execute()
    with pytest.raises(AssertionError):
        yield from tr.execute()
    with pytest.raises(AssertionError):
        yield from tr.incr('foo')


@pytest.mark.run_loop
def test_connection_closed(redis):
    tr = redis.multi_exec()
    fut1 = tr.quit()
    fut2 = tr.incrby('foo', 1.0)
    fut3 = tr.connection.execute('INCRBY', 'foo', '1.0')
    with pytest.raises(MultiExecError):
        yield from tr.execute()

    assert fut1.done() is True
    assert fut2.done() is True
    assert fut3.done() is True

    try:
        assert (yield from fut1) == b'OK'
    except asyncio.CancelledError:
        pass
    assert fut2.cancelled() is False
    assert isinstance(fut2.exception(), TypeError)

    assert fut3.cancelled() is True


@pytest.mark.run_loop
def test_discard(redis):
    yield from redis.delete('foo')
    tr = redis.multi_exec()
    fut1 = tr.incrby('foo', 1.0)
    fut2 = tr.connection.execute('MULTI')
    fut3 = tr.connection.execute('incr', 'foo')

    with pytest.raises(ReplyError):
        yield from tr.execute()
    with pytest.raises(TypeError):
        yield from fut1
    with pytest.raises(ReplyError):
        yield from fut2
    # with pytest.raises(ReplyError):
    res = yield from fut3
    assert res == 1


@pytest.mark.run_loop
def test_exec_error(redis):
    tr = redis.multi_exec()
    fut = tr.connection.execute('INCRBY', 'key', '1.0')
    with pytest.raises(MultiExecError):
        yield from tr.execute()
    with pytest.raises(ReplyError):
        yield from fut

    yield from redis.set('foo', 'bar')
    tr = redis.multi_exec()
    fut = tr.incrbyfloat('foo', 1.1)
    res = yield from tr.execute(return_exceptions=True)
    assert isinstance(res[0], ReplyError)
    with pytest.raises(ReplyError):
        yield from fut


@pytest.mark.run_loop
def test_command_errors(redis):
    tr = redis.multi_exec()
    fut = tr.incrby('key', 1.0)
    with pytest.raises(MultiExecError):
        yield from tr.execute()
    with pytest.raises(TypeError):
        yield from fut


@pytest.mark.run_loop
def test_several_command_errors(redis):
    tr = redis.multi_exec()
    fut1 = tr.incrby('key', 1.0)
    fut2 = tr.rename('bar', 'bar')
    with pytest.raises(MultiExecError):
        yield from tr.execute()
    with pytest.raises(TypeError):
        yield from fut1
    with pytest.raises(ValueError):
        yield from fut2


@pytest.mark.run_loop
def test_error_in_connection(redis):
    yield from redis.set('foo', 1)
    tr = redis.multi_exec()
    fut1 = tr.mget('foo', None)
    fut2 = tr.incr('foo')
    with pytest.raises(MultiExecError):
        yield from tr.execute()
    with pytest.raises(TypeError):
        yield from fut1
    yield from fut2


@pytest.mark.run_loop
def test_watch_unwatch(redis):
    res = yield from redis.watch('key')
    assert res is True
    res = yield from redis.watch('key', 'key')
    assert res is True

    with pytest.raises(TypeError):
        yield from redis.watch(None)
    with pytest.raises(TypeError):
        yield from redis.watch('key', None)
    with pytest.raises(TypeError):
        yield from redis.watch('key', 'key', None)

    res = yield from redis.unwatch()
    assert res is True


@pytest.mark.run_loop
def test_encoding(redis):
    res = yield from redis.set('key', 'value')
    assert res is True
    res = yield from redis.hmset(
        'hash-key', 'foo', 'val1', 'bar', 'val2')
    assert res is True

    tr = redis.multi_exec()
    fut1 = tr.get('key')
    fut2 = tr.get('key', encoding='utf-8')
    fut3 = tr.hgetall('hash-key', encoding='utf-8')
    yield from tr.execute()
    res = yield from fut1
    assert res == b'value'
    res = yield from fut2
    assert res == 'value'
    res = yield from fut3
    assert res == {'foo': 'val1', 'bar': 'val2'}


@pytest.mark.run_loop
def test_global_encoding(redis, create_redis, server, loop):
    redis = yield from create_redis(
        server.tcp_address,
        loop=loop, encoding='utf-8')
    res = yield from redis.set('key', 'value')
    assert res is True
    res = yield from redis.hmset(
        'hash-key', 'foo', 'val1', 'bar', 'val2')
    assert res is True

    tr = redis.multi_exec()
    fut1 = tr.get('key')
    fut2 = tr.get('key', encoding='utf-8')
    fut3 = tr.hgetall('hash-key', encoding='utf-8')
    yield from tr.execute()
    res = yield from fut1
    assert res == 'value'
    res = yield from fut2
    assert res == 'value'
    res = yield from fut3
    assert res == {'foo': 'val1', 'bar': 'val2'}


@pytest.mark.run_loop
def test_transaction__watch_error(redis, create_redis, server, loop):
    other = yield from create_redis(
        server.tcp_address, loop=loop)

    ok = yield from redis.set('foo', 'bar')
    assert ok is True

    ok = yield from redis.watch('foo')
    assert ok is True

    ok = yield from other.set('foo', 'baz')
    assert ok is True

    tr = redis.multi_exec()
    fut1 = tr.set('foo', 'foo')
    fut2 = tr.get('bar')
    with pytest.raises(MultiExecError):
        yield from tr.execute()
    with pytest.raises(WatchVariableError):
        yield from fut1
    with pytest.raises(WatchVariableError):
        yield from fut2


@pytest.mark.run_loop
def test_multi_exec_and_pool_release(redis):
    # Test the case when pool connection is released before
    # `exec` result is received.

    slow_script = """
    local a = tonumber(redis.call('time')[1])
    local b = a + 1
    while (a < b)
    do
        a = tonumber(redis.call('time')[1])
    end
    """

    tr = redis.multi_exec()
    fut1 = tr.eval(slow_script)
    ret, = yield from tr.execute()
    assert ret is None
    assert (yield from fut1) is None



import asyncio

from ..util import wait_ok, wait_convert
from ..commands import Redis
from .pool import create_sentinel_pool


@asyncio.coroutine
def create_sentinel(sentinels, *, db=None, password=None,
                    encoding=None, minsize=1, maxsize=10,
                    ssl=None, timeout=0.2, loop=None):
    """Creates Redis Sentinel client.

    `sentinels` is a list of sentinel nodes.
    """

    if loop is None:
        loop = asyncio.get_event_loop()

    pool = yield from create_sentinel_pool(sentinels,
                                           db=db,
                                           password=password,
                                           encoding=encoding,
                                           minsize=minsize,
                                           maxsize=maxsize,
                                           ssl=ssl,
                                           timeout=timeout,
                                           loop=loop)
    return RedisSentinel(pool)


class RedisSentinel:
    """Redis Sentinel client."""

    def __init__(self, pool):
        self._pool = pool

    def close(self):
        """Close client connections."""
        self._pool.close()

    @asyncio.coroutine
    def wait_closed(self):
        """Coroutine waiting until underlying connections are closed."""
        yield from self._pool.wait_closed()

    @property
    def closed(self):
        """True if connection is closed."""
        return self._pool.closed

    def master_for(self, name):
        """Returns Redis client to master Redis server."""
        # TODO: make class configurable
        return Redis(self._pool.master_for(name))

    def slave_for(self, name):
        """Returns Redis client to slave Redis server."""
        # TODO: make class configurable
        return Redis(self._pool.slave_for(name))

    def execute(self, command, *args, **kwargs):
        """Execute Sentinel command.

        It will be prefixed with SENTINEL automatically.
        """
        return self._pool.execute(
            b'SENTINEL', command, *args, **kwargs)

    def ping(self):
        """Send PING command to Sentinel instance(s)."""
        # TODO: add kwargs allowing to pick sentinel to send command to.
        return (yield from self._pool.execute(b'PING'))

    def master(self, name):
        """Returns a dictionary containing the specified masters state."""
        fut = self.execute(b'MASTER', name, encoding='utf-8')
        return wait_convert(fut, parse_sentinel_master)

    def master_address(self, name):
        """Returns a (host, port) pair for the given ``name``."""
        fut = self.execute(b'get-master-addr-by-name', name, encoding='utf-8')
        return wait_convert(fut, parse_address)

    def masters(self):
        """Returns a list of dictionaries containing each master's state."""
        fut = self.execute(b'MASTERS', encoding='utf-8')
        # TODO: process masters: we can adjust internal state
        return wait_convert(fut, parse_sentinel_masters)

    def slaves(self, name):
        """Returns a list of slaves for ``name``."""
        fut = self.execute(b'SLAVES', name, encoding='utf-8')
        return wait_convert(fut, parse_sentinel_slaves_and_sentinels)

    def sentinels(self, name):
        """Returns a list of sentinels for ``name``."""
        fut = self.execute(b'SENTINELS', name, encoding='utf-8')
        return wait_convert(fut, parse_sentinel_slaves_and_sentinels)

    def monitor(self, name, ip, port, quorum):
        """Add a new master to Sentinel to be monitored."""
        fut = self.execute(b'MONITOR', name, ip, port, quorum)
        return wait_ok(fut)

    def remove(self, name):
        """Remove a master from Sentinel's monitoring."""
        fut = self.execute(b'REMOVE', name)
        return wait_ok(fut)

    def set(self, name, option, value):
        """Set Sentinel monitoring parameters for a given master."""
        fut = self.execute(b"SET", name, option, value)
        return wait_ok(fut)

    def failover(self, name):
        """Force a failover of a named master."""
        fut = self.execute(b'FAILOVER', name)
        return wait_ok(fut)

    def check_quorum(self, name):
        """
        Check if the current Sentinel configuration is able
        to reach the quorum needed to failover a master,
        and the majority needed to authorize the failover.
        """
        return self.execute(b'CKQUORUM', name)


SENTINEL_STATE_TYPES = {
    'can-failover-its-master': int,
    'config-epoch': int,
    'down-after-milliseconds': int,
    'failover-timeout': int,
    'info-refresh': int,
    'last-hello-message': int,
    'last-ok-ping-reply': int,
    'last-ping-reply': int,
    'last-ping-sent': int,
    'master-link-down-time': int,
    'master-port': int,
    'num-other-sentinels': int,
    'num-slaves': int,
    'o-down-time': int,
    'pending-commands': int,
    'link-pending-commands': int,
    'link-refcount': int,
    'parallel-syncs': int,
    'port': int,
    'quorum': int,
    'role-reported-time': int,
    's-down-time': int,
    'slave-priority': int,
    'slave-repl-offset': int,
    'voted-leader-epoch': int,
    'flags': lambda s: frozenset(s.split(',')),  # TODO: make flags enum?
}


def pairs_to_dict_typed(response, type_info):
    it = iter(response)
    result = {}
    for key, value in zip(it, it):
        if key in type_info:
            try:
                value = type_info[key](value)
            except (TypeError, ValueError):
                # if for some reason the value can't be coerced, just use
                # the string value
                pass
        result[key] = value
    return result


def parse_sentinel_masters(response):
    result = {}
    for item in response:
        state = pairs_to_dict_typed(item, SENTINEL_STATE_TYPES)
        result[state['name']] = state
    return result


def parse_sentinel_slaves_and_sentinels(response):
    return [pairs_to_dict_typed(item, SENTINEL_STATE_TYPES)
            for item in response]


def parse_sentinel_master(response):
    return pairs_to_dict_typed(response, SENTINEL_STATE_TYPES)


def parse_address(value):
    if value is not None:
        return (value[0], int(value[1]))



import asyncio
import pytest

from aioredis.util import create_future, async_task


@asyncio.coroutine
def _reader(channel, output, waiter, conn):
    yield from conn.execute('subscribe', channel)
    ch = conn.pubsub_channels[channel]
    waiter.set_result(conn)
    while (yield from ch.wait_message()):
        msg = yield from ch.get()
        yield from output.put(msg)


@pytest.mark.run_loop
def test_publish(create_connection, redis, server, loop):
    out = asyncio.Queue(loop=loop)
    fut = create_future(loop=loop)
    conn = yield from create_connection(
        server.tcp_address, loop=loop)
    sub = async_task(_reader('chan:1', out, fut, conn), loop=loop)

    yield from fut
    yield from redis.publish('chan:1', 'Hello')
    msg = yield from out.get()
    assert msg == b'Hello'

    sub.cancel()


@pytest.mark.run_loop
def test_publish_json(create_connection, redis, server, loop):
    out = asyncio.Queue(loop=loop)
    fut = create_future(loop=loop)
    conn = yield from create_connection(
        server.tcp_address, loop=loop)
    sub = async_task(_reader('chan:1', out, fut, conn), loop=loop)

    yield from fut

    res = yield from redis.publish_json('chan:1', {"Hello": "world"})
    assert res == 1    # recievers

    msg = yield from out.get()
    assert msg == b'{"Hello": "world"}'
    sub.cancel()


@pytest.mark.run_loop
def test_subscribe(redis):
    res = yield from redis.subscribe('chan:1', 'chan:2')
    assert redis.in_pubsub == 2

    ch1 = redis.channels['chan:1']
    ch2 = redis.channels['chan:2']

    assert res == [ch1, ch2]
    assert ch1.is_pattern is False
    assert ch2.is_pattern is False

    res = yield from redis.unsubscribe('chan:1', 'chan:2')
    assert res == [[b'unsubscribe', b'chan:1', 1],
                   [b'unsubscribe', b'chan:2', 0]]


@pytest.mark.run_loop
def test_psubscribe(redis, create_redis, server, loop):
    sub = redis
    res = yield from sub.psubscribe('patt:*', 'chan:*')
    assert sub.in_pubsub == 2

    pat1 = sub.patterns['patt:*']
    pat2 = sub.patterns['chan:*']
    assert res == [pat1, pat2]

    pub = yield from create_redis(
        server.tcp_address, loop=loop)
    yield from pub.publish_json('chan:123', {"Hello": "World"})
    res = yield from pat2.get_json()
    assert res == (b'chan:123', {"Hello": "World"})

    res = yield from sub.punsubscribe('patt:*', 'patt:*', 'chan:*')
    assert res == [[b'punsubscribe', b'patt:*', 1],
                   [b'punsubscribe', b'patt:*', 1],
                   [b'punsubscribe', b'chan:*', 0],
                   ]


@pytest.redis_version(
    2, 8, 0, reason='PUBSUB CHANNELS is available since redis>=2.8.0')
@pytest.mark.run_loop
def test_pubsub_channels(create_redis, server, loop):
    redis = yield from create_redis(
        server.tcp_address, loop=loop)
    res = yield from redis.pubsub_channels()
    assert res == []

    res = yield from redis.pubsub_channels('chan:*')
    assert res == []

    sub = yield from create_redis(
        server.tcp_address, loop=loop)
    yield from sub.subscribe('chan:1')

    res = yield from redis.pubsub_channels()
    assert res == [b'chan:1']

    res = yield from redis.pubsub_channels('ch*')
    assert res == [b'chan:1']

    yield from sub.unsubscribe('chan:1')
    yield from sub.psubscribe('chan:*')

    res = yield from redis.pubsub_channels()
    assert res == []


@pytest.redis_version(
    2, 8, 0, reason='PUBSUB NUMSUB is available since redis>=2.8.0')
@pytest.mark.run_loop
def test_pubsub_numsub(create_redis, server, loop):
    redis = yield from create_redis(
        server.tcp_address, loop=loop)
    res = yield from redis.pubsub_numsub()
    assert res == {}

    res = yield from redis.pubsub_numsub('chan:1')
    assert res == {b'chan:1': 0}

    sub = yield from create_redis(
        server.tcp_address, loop=loop)
    yield from sub.subscribe('chan:1')

    res = yield from redis.pubsub_numsub()
    assert res == {}

    res = yield from redis.pubsub_numsub('chan:1')
    assert res == {b'chan:1': 1}

    res = yield from redis.pubsub_numsub('chan:2')
    assert res == {b'chan:2': 0}

    res = yield from redis.pubsub_numsub('chan:1', 'chan:2')
    assert res == {b'chan:1': 1, b'chan:2': 0}

    yield from sub.unsubscribe('chan:1')
    yield from sub.psubscribe('chan:*')

    res = yield from redis.pubsub_numsub()
    assert res == {}


@pytest.redis_version(
    2, 8, 0, reason='PUBSUB NUMPAT is available since redis>=2.8.0')
@pytest.mark.run_loop
def test_pubsub_numpat(create_redis, server, loop, redis):
    sub = yield from create_redis(
        server.tcp_address, loop=loop)

    res = yield from redis.pubsub_numpat()
    assert res == 0

    yield from sub.subscribe('chan:1')
    res = yield from redis.pubsub_numpat()
    assert res == 0

    yield from sub.psubscribe('chan:*')
    res = yield from redis.pubsub_numpat()
    assert res == 1


@pytest.mark.run_loop
def test_close_pubsub_channels(redis, loop):
    ch, = yield from redis.subscribe('chan:1')

    @asyncio.coroutine
    def waiter(ch):
        assert not (yield from ch.wait_message())

    tsk = async_task(waiter(ch), loop=loop)
    redis.close()
    yield from redis.wait_closed()
    yield from tsk


@pytest.mark.run_loop
def test_close_pubsub_patterns(redis, loop):
    ch, = yield from redis.psubscribe('chan:*')

    @asyncio.coroutine
    def waiter(ch):
        assert not (yield from ch.wait_message())

    tsk = async_task(waiter(ch), loop=loop)
    redis.close()
    yield from redis.wait_closed()
    yield from tsk


@pytest.mark.run_loop
def test_close_cancelled_pubsub_channel(redis, loop):
    ch, = yield from redis.subscribe('chan:1')

    @asyncio.coroutine
    def waiter(ch):
        with pytest.raises(asyncio.CancelledError):
            yield from ch.wait_message()

    tsk = async_task(waiter(ch), loop=loop)
    yield from asyncio.sleep(0, loop=loop)
    tsk.cancel()


@pytest.mark.run_loop
def test_channel_get_after_close(create_redis, loop, server):
    sub = yield from create_redis(
        server.tcp_address, loop=loop)
    pub = yield from create_redis(
        server.tcp_address, loop=loop)
    ch, = yield from sub.subscribe('chan:1')

    @asyncio.coroutine
    def waiter():
        while True:
            msg = yield from ch.get()
            if msg is None:
                break
            assert msg == b'message'

    tsk = async_task(waiter(), loop=loop)

    yield from pub.publish('chan:1', 'message')
    sub.close()
    yield from tsk


@pytest.mark.run_loop
def test_subscribe_concurrency(create_redis, server, loop):
    sub = yield from create_redis(
        server.tcp_address, loop=loop)
    pub = yield from create_redis(
        server.tcp_address, loop=loop)

    @asyncio.coroutine
    def subscribe(*args):
        return (yield from sub.subscribe(*args))

    @asyncio.coroutine
    def publish(*args):
        yield from asyncio.sleep(0, loop=loop)
        return (yield from pub.publish(*args))

    res = yield from asyncio.gather(
        subscribe('channel:0'),
        publish('channel:0', 'Hello'),
        subscribe('channel:1'),
        loop=loop)
    (ch1,), subs, (ch2,) = res

    assert ch1.name == b'channel:0'
    assert subs == 1
    assert ch2.name == b'channel:1'



import asyncio
import pytest

from aioredis.util import async_task
from aioredis.locks import Lock


@pytest.mark.run_loop
def test_finished_waiter_cancelled(loop):
    lock = Lock(loop=loop)

    ta = async_task(lock.acquire(), loop=loop)
    yield from asyncio.sleep(0, loop=loop)
    assert lock.locked()

    tb = async_task(lock.acquire(), loop=loop)
    yield from asyncio.sleep(0, loop=loop)
    assert len(lock._waiters) == 1

    # Create a second waiter, wake up the first, and cancel it.
    # Without the fix, the second was not woken up and the lock
    # will never be locked
    async_task(lock.acquire(), loop=loop)
    yield from asyncio.sleep(0, loop=loop)
    lock.release()
    tb.cancel()

    yield from asyncio.sleep(0, loop=loop)
    assert ta.done()
    assert tb.cancelled()

    yield from asyncio.sleep(0, loop=loop)
    assert lock.locked()



import asyncio
import aioredis


async def main():

    redis = await aioredis.create_redis(
        ('localhost', 6379))

    await redis.delete('something:hash',
                       'something:set',
                       'something:zset')
    await redis.mset('something', 'value',
                     'something:else', 'else')
    await redis.hmset('something:hash',
                      'something:1', 'value:1',
                      'something:2', 'value:2')
    await redis.sadd('something:set', 'something:1',
                     'something:2', 'something:else')
    await redis.zadd('something:zset', 1, 'something:1',
                     2, 'something:2', 3, 'something:else')

    await go(redis)
    redis.close()
    await redis.wait_closed()


async def go(redis):
    async for key in redis.iscan(match='something*'):
        print('Matched:', key)

    key = 'something:hash'

    async for name, val in redis.ihscan(key, match='something*'):
        print('Matched:', name, '->', val)

    key = 'something:set'

    async for val in redis.isscan(key, match='something*'):
        print('Matched:', val)

    key = 'something:zset'

    async for val, score in redis.izscan(key, match='something*'):
        print('Matched:', val, ':', score)


if __name__ == '__main__':
    import os
    if 'redis_version:2.6' not in os.environ.get('REDIS_VERSION', ''):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(main())



__all__ = [
    'RedisError',
    'ProtocolError',
    'ReplyError',
    'PipelineError',
    'MultiExecError',
    'WatchVariableError',
    'ChannelClosedError',
    'ConnectionClosedError',
    'PoolClosedError',
    'MasterNotFoundError',
    'SlaveNotFoundError',
    'ReadOnlyError',
    ]


class RedisError(Exception):
    """Base exception class for aioredis exceptions."""


class ProtocolError(RedisError):
    """Raised when protocol error occurs."""


class ReplyError(RedisError):
    """Raised for redis error replies (-ERR)."""


class PipelineError(ReplyError):
    """Raised if command within pipeline raised error."""

    def __init__(self, errors):
        super().__init__('{} errors:'.format(self.__class__.__name__), errors)


class MultiExecError(PipelineError):
    """Raised if command within MULTI/EXEC block caused error."""


class WatchVariableError(MultiExecError):
    """Raised if watched variable changed (EXEC returns None)."""


class ChannelClosedError(RedisError):
    """Raised when Pub/Sub channel is unsubscribed and messages queue is empty.
    """


class ReadOnlyError(RedisError):
    """Raised from slave when read-only mode is enabled"""


class MasterNotFoundError(RedisError):
    """Raised for sentinel master not found error."""


class SlaveNotFoundError(RedisError):
    """Raised for sentinel slave not found error."""


class ConnectionClosedError(RedisError):
    """Raised if connection to server was closed."""


class PoolClosedError(RedisError):
    """Raised if pool is closed."""



import pytest

from aioredis.errors import ProtocolError, ReplyError
from aioredis.parser import PyReader


@pytest.fixture
def reader():
    return PyReader()


def test_nothing(reader):
    assert reader.gets() is False


def test_error_when_feeding_non_string(reader):
    with pytest.raises(TypeError):
        reader.feed(1)


@pytest.mark.parametrize('data', [
    b'x', b'$5\r\nHello world',
    b':None\r\n', b':1.2\r\n', b':1,2\r\n',
], ids=[
    'Bad control char',
    'Invalid bulk length',
    'Invalid int - none',
    'Invalid int - dot',
    'Invalid int - comma',
])
def test_protocol_error(reader, data):
    reader.feed(data)
    with pytest.raises(ProtocolError):
        reader.gets()
    # not functional any more
    with pytest.raises(ProtocolError):
        reader.gets()


class CustomExc(Exception):
    pass


@pytest.mark.parametrize('exc,arg', [
    (RuntimeError, RuntimeError),
    (CustomExc, lambda e: CustomExc(e)),
    ], ids=['RuntimeError', 'callable'])
def test_protocol_error_with_custom_class(exc, arg):
    reader = PyReader(protocolError=arg)
    reader.feed(b"x")
    with pytest.raises(exc):
        reader.gets()


@pytest.mark.parametrize('init', [
    dict(protocolError="wrong"),
    dict(replyError="wrong"),
], ids=['wrong protocolError', 'wrong replyError'])
def test_fail_with_wrong_error_class(init):
    with pytest.raises(TypeError):
        PyReader(**init)


def test_error_string(reader):
    reader.feed(b"-error\r\n")
    error = reader.gets()

    assert isinstance(error, ReplyError)
    assert error.args == ("error",)


@pytest.mark.parametrize('exc,arg', [
    (RuntimeError, RuntimeError),
    (CustomExc, lambda e: CustomExc(e)),
    ], ids=['RuntimeError', 'callable'])
def test_error_string_with_custom_class(exc, arg):
    reader = PyReader(replyError=arg)
    reader.feed(b"-error\r\n")
    error = reader.gets()

    assert isinstance(error, exc)
    assert error.args == ("error",)


def test_errors_in_nested_multi_bulk(reader):
    reader.feed(b"*2\r\n-err0\r\n-err1\r\n")

    for r, error in zip(("err0", "err1"), reader.gets()):
        assert isinstance(error, ReplyError)
        assert error.args == (r,)


def test_integer(reader):
    value = 2**63-1  # Largest 64-bit signed integer
    reader.feed((":%d\r\n" % value).encode("ascii"))
    assert reader.gets() == value


def test_status_string(reader):
    reader.feed(b"+ok\r\n")
    assert reader.gets() == b"ok"


@pytest.mark.parametrize('data,expected', [
    (b'$0\r\n\r\n', b''),
    (b'$-1\r\n', None),
    (b'$5\r\nhello\r\n', b'hello'),
], ids=['Empty', 'null', 'hello'])
def test_bulk_string(reader, data, expected):
    reader.feed(data)
    assert reader.gets() == expected


def test_bulk_string_without_encoding(reader):
    snowman = b"\xe2\x98\x83"
    reader.feed(b"$3\r\n" + snowman + b"\r\n")
    assert reader.gets() == snowman


@pytest.mark.parametrize('encoding,expected', [
    ('utf-8', b"\xe2\x98\x83".decode('utf-8')),
    ('utf-32', b"\xe2\x98\x83"),
], ids=['utf-8', 'utf-32'])
def test_bulk_string_with_encoding(encoding, expected):
    snowman = b"\xe2\x98\x83"
    reader = PyReader(encoding=encoding)
    reader.feed(b"$3\r\n" + snowman + b"\r\n")
    assert reader.gets() == expected


def test_bulk_string_with_invalid_encoding():
    reader = PyReader(encoding="unknown")
    reader.feed(b"$5\r\nhello\r\n")
    with pytest.raises(LookupError):
        reader.gets()


def test_bulk_string_wait_buffer(reader):
    reader.feed(b'$5\r\nH')
    assert not reader.gets()
    reader.feed(b'ello')
    assert not reader.gets()
    reader.feed(b'\r\n')
    assert reader.gets() == b'Hello'


@pytest.mark.parametrize('data,expected', [
    (b"*-1\r\n", None),
    (b"*0\r\n", []),
    (b"*2\r\n$5\r\nhello\r\n$5\r\nworld\r\n", [b'hello', b'world']),
], ids=['Null', 'Empty list', 'hello world'])
def test_null_multi_bulk(reader, data, expected):
    reader.feed(data)
    assert reader.gets() == expected


@pytest.mark.parametrize('data', [
    (b"*2\r\n$5\r\nhello\r\n", b':1'),
    (b'*2\r\n:1\r\n*1\r\n', b'+hello'),
    (b'*2\r\n+hello\r\n+world',),
    (b'*2\r\n*1\r\n+hello\r\n*1\r\n+world',),
], ids=['First in bulk',
        'Error in nested',
        'Multiple errors',
        'Multiple nested'])
def test_multi_bulk_with_invalid_encoding_and_partial_reply(data):
    reader = PyReader(encoding="unknown")
    for chunk in data:
        reader.feed(chunk)
        assert reader.gets() is False
    reader.feed(b"\r\n")
    with pytest.raises(LookupError):
        reader.gets()

    reader.feed(b':1\r\n')
    assert reader.gets() == 1


def test_nested_multi_bulk(reader):
    reader.feed(b"*2\r\n*2\r\n$5\r\nhello\r\n$5\r\nworld\r\n$1\r\n!\r\n")
    assert reader.gets() == [[b"hello", b"world"], b"!"]


def test_nested_multi_bulk_depth(reader):
    reader.feed(b"*1\r\n*1\r\n*1\r\n*1\r\n$1\r\n!\r\n")
    assert reader.gets() == [[[[b"!"]]]]


@pytest.mark.parametrize('encoding,expected', [
    ('utf-8', b"\xe2\x98\x83".decode('utf-8')),
    ('utf-32', b"\xe2\x98\x83"),
], ids=['utf-8', 'utf-32'])
def test_simple_string_with_encoding(encoding, expected):
    snowman = b"\xe2\x98\x83"
    reader = PyReader(encoding=encoding)
    reader.feed(b"+" + snowman + b"\r\n")
    assert reader.gets() == expected


def test_invalid_offset(reader):
    data = b"+ok\r\n"
    with pytest.raises(ValueError):
        reader.feed(data, 6)


def test_invalid_length(reader):
    data = b"+ok\r\n"
    with pytest.raises(ValueError):
        reader.feed(data, 0, 6)


def test_ok_offset(reader):
    data = b"blah+ok\r\n"
    reader.feed(data, 4)
    assert reader.gets() == b"ok"


def test_ok_length(reader):
    data = b"blah+ok\r\n"
    reader.feed(data, 4, len(data)-4)
    assert reader.gets() == b"ok"


@pytest.mark.xfail()
def test_maxbuf(reader):
    defaultmaxbuf = reader.getmaxbuf()
    reader.setmaxbuf(0)
    assert 0 == reader.getmaxbuf()
    reader.setmaxbuf(10000)
    assert 10000 == reader.getmaxbuf()
    reader.setmaxbuf(None)
    assert defaultmaxbuf == reader.getmaxbuf()
    with pytest.raises(ValueError):
        reader.setmaxbuf(-4)



import asyncio
import aioredis


STOPWORD = 'STOP'


@asyncio.coroutine
def pubsub():
    pool = yield from aioredis.create_pool(
        ('localhost', 6379),
        minsize=5, maxsize=10)

    @asyncio.coroutine
    def reader(channel):
        while (yield from channel.wait_message()):
            msg = yield from channel.get(encoding='utf-8')
            # ... process message ...
            print("message in {}: {}".format(channel.name, msg))

            if msg == STOPWORD:
                return

    with (yield from pool) as conn:
        raw_result = yield from conn.execute_pubsub('subscribe', 'channel:1')
        print('raw result:', raw_result)
        channel = conn.pubsub_channels['channel:1']
        yield from reader(channel)  # wait for reader to complete
        yield from conn.execute_pubsub('unsubscribe', 'channel:1')

    # Explicit connection usage
    conn = yield from pool.acquire()
    try:
        raw_result = yield from conn.execute_pubsub('subscribe', 'channel:1')
        print('raw result:', raw_result)
        channel = conn.pubsub_channels['channel:1']
        yield from reader(channel)  # wait for reader to complete
        yield from conn.execute_pubsub('unsubscribe', 'channel:1')
    finally:
        pool.release(conn)

    pool.close()
    yield from pool.wait_closed()    # closing all open connections


def main():
    loop = asyncio.get_event_loop()
    tsk = asyncio.async(pubsub(), loop=loop)

    @asyncio.coroutine
    def publish():
        pub = yield from aioredis.create_redis(
            ('localhost', 6379))
        while not tsk.done():
            # wait for clients to subscribe
            while True:
                subs = yield from pub.pubsub_numsub('channel:1')
                if subs[b'channel:1'] == 1:
                    break
                yield from asyncio.sleep(0, loop=loop)
            # publish some messages
            for msg in ['one', 'two', 'three']:
                yield from pub.publish('channel:1', msg)
            # send stop word
            yield from pub.publish('channel:1', STOPWORD)
        pub.close()
        yield from pub.wait_closed()

    loop.run_until_complete(asyncio.gather(publish(), tsk, loop=loop))


if __name__ == '__main__':
    import os
    if 'redis_version:2.6' not in os.environ.get('REDIS_VERSION', ''):
        main()



import asyncio
import pytest


@pytest.mark.run_loop
async def test_pubsub_channel_iter(create_redis, server, loop):
    sub = await create_redis(server.tcp_address, loop=loop)
    pub = await create_redis(server.tcp_address, loop=loop)

    ch, = await sub.subscribe('chan:1')

    async def coro(ch):
        lst = []
        async for msg in ch.iter():
            lst.append(msg)
        return lst

    tsk = asyncio.ensure_future(coro(ch), loop=loop)
    await pub.publish_json('chan:1', {'Hello': 'World'})
    await pub.publish_json('chan:1', ['message'])
    await asyncio.sleep(0, loop=loop)
    ch.close()
    assert await tsk == [b'{"Hello": "World"}', b'["message"]']



import pytest
import asyncio
import json
import sys

from unittest import mock

from aioredis import ChannelClosedError
from aioredis.abc import AbcChannel
from aioredis.pubsub import Receiver, _Sender
from aioredis.util import async_task


def test_listener_channel(loop):
    mpsc = Receiver(loop=loop)
    assert not mpsc.is_active

    ch_a = mpsc.channel("channel:1")
    assert isinstance(ch_a, AbcChannel)
    assert mpsc.is_active

    ch_b = mpsc.channel('channel:1')
    assert ch_a is ch_b
    assert ch_a.name == ch_b.name
    assert ch_a.is_pattern == ch_b.is_pattern
    assert mpsc.is_active

    # remember id; drop refs to objects and create new one;
    ch_a.close()
    assert not ch_a.is_active

    assert not mpsc.is_active
    ch = mpsc.channel("channel:1")
    assert ch is not ch_a

    assert dict(mpsc.channels) == {b'channel:1': ch}
    assert dict(mpsc.patterns) == {}


def test_listener_pattern(loop):
    mpsc = Receiver(loop=loop)
    assert not mpsc.is_active

    ch_a = mpsc.pattern("*")
    assert isinstance(ch_a, AbcChannel)
    assert mpsc.is_active

    ch_b = mpsc.pattern('*')
    assert ch_a is ch_b
    assert ch_a.name == ch_b.name
    assert ch_a.is_pattern == ch_b.is_pattern
    assert mpsc.is_active

    # remember id; drop refs to objects and create new one;
    ch_a.close()
    assert not ch_a.is_active

    assert not mpsc.is_active
    ch = mpsc.pattern("*")
    assert ch is not ch_a

    assert dict(mpsc.channels) == {}
    assert dict(mpsc.patterns) == {b'*': ch}


@pytest.mark.run_loop
def test_sender(loop):
    receiver = mock.Mock()

    sender = _Sender(receiver, 'name', is_pattern=False, loop=loop)
    assert isinstance(sender, AbcChannel)
    assert sender.name == b'name'
    assert sender.is_pattern is False
    assert sender.is_active is True

    with pytest.raises(RuntimeError):
        yield from sender.get()
    assert receiver.mock_calls == []

    sender.put_nowait(b'some data')
    assert receiver.mock_calls == [
        mock.call._put_nowait(b'some data', sender=sender),
        ]


def test_sender_close():
    receiver = mock.Mock()
    loop = mock.Mock()
    sender = _Sender(receiver, 'name', is_pattern=False, loop=loop)
    sender.close()
    assert receiver.mock_calls == [mock.call._close(sender)]
    sender.close()
    assert receiver.mock_calls == [mock.call._close(sender)]
    receiver.reset_mock()
    assert receiver.mock_calls == []
    sender.close()
    assert receiver.mock_calls == []


@pytest.mark.run_loop
def test_subscriptions(create_connection, server, loop):
    sub = yield from create_connection(server.tcp_address, loop=loop)
    pub = yield from create_connection(server.tcp_address, loop=loop)

    mpsc = Receiver(loop=loop)
    yield from sub.execute_pubsub('subscribe',
                                  mpsc.channel('channel:1'),
                                  mpsc.channel('channel:3'))
    res = yield from pub.execute("publish", "channel:3", "Hello world")
    assert res == 1
    res = yield from pub.execute("publish", "channel:1", "Hello world")
    assert res == 1
    assert mpsc.is_active

    ch, msg = yield from mpsc.get()
    assert ch.name == b'channel:3'
    assert not ch.is_pattern
    assert msg == b"Hello world"

    ch, msg = yield from mpsc.get()
    assert ch.name == b'channel:1'
    assert not ch.is_pattern
    assert msg == b"Hello world"


@pytest.mark.run_loop
def test_unsubscribe(create_connection, server, loop):
    sub = yield from create_connection(server.tcp_address, loop=loop)
    pub = yield from create_connection(server.tcp_address, loop=loop)

    mpsc = Receiver(loop=loop)
    yield from sub.execute_pubsub('subscribe',
                                  mpsc.channel('channel:1'),
                                  mpsc.channel('channel:3'))
    res = yield from pub.execute("publish", "channel:3", "Hello world")
    assert res == 1
    res = yield from pub.execute("publish", "channel:1", "Hello world")
    assert res == 1
    assert mpsc.is_active

    assert (yield from mpsc.wait_message()) is True
    ch, msg = yield from mpsc.get()
    assert ch.name == b'channel:3'
    assert not ch.is_pattern
    assert msg == b"Hello world"

    assert (yield from mpsc.wait_message()) is True
    ch, msg = yield from mpsc.get()
    assert ch.name == b'channel:1'
    assert not ch.is_pattern
    assert msg == b"Hello world"

    yield from sub.execute_pubsub('unsubscribe', 'channel:1')
    assert mpsc.is_active

    res = yield from pub.execute("publish", "channel:3", "message")
    assert res == 1
    assert (yield from mpsc.wait_message()) is True
    ch, msg = yield from mpsc.get()
    assert ch.name == b'channel:3'
    assert not ch.is_pattern
    assert msg == b"message"

    yield from sub.execute_pubsub('unsubscribe', 'channel:3')
    assert not mpsc.is_active
    res = yield from mpsc.get()
    assert res is None


@pytest.mark.run_loop
def test_stopped(create_connection, server, loop):
    sub = yield from create_connection(server.tcp_address, loop=loop)
    pub = yield from create_connection(server.tcp_address, loop=loop)

    mpsc = Receiver(loop=loop)
    yield from sub.execute_pubsub('subscribe', mpsc.channel('channel:1'))
    assert mpsc.is_active
    mpsc.stop()

    with pytest.logs('aioredis', 'DEBUG') as cm:
        yield from pub.execute('publish', 'channel:1', b'Hello')
        yield from asyncio.sleep(0, loop=loop)

    assert len(cm.output) == 1
    # Receiver must have 1 EndOfStream message
    warn_messaege = (
        "WARNING:aioredis:Pub/Sub listener message after stop: "
        "sender: <_Sender name:b'channel:1', is_pattern:False, receiver:"
        "<Receiver is_active:True, senders:1, qsize:1>>, data: b'Hello'"
    )
    assert cm.output == [warn_messaege]

    assert (yield from mpsc.get()) is None
    with pytest.raises(ChannelClosedError):
        yield from mpsc.get()
    res = yield from mpsc.wait_message()
    assert res is False


@pytest.mark.run_loop
def test_wait_message(create_connection, server, loop):
    sub = yield from create_connection(server.tcp_address, loop=loop)
    pub = yield from create_connection(server.tcp_address, loop=loop)

    mpsc = Receiver(loop=loop)
    yield from sub.execute_pubsub('subscribe', mpsc.channel('channel:1'))
    fut = async_task(mpsc.wait_message(), loop=loop)
    assert not fut.done()
    yield from asyncio.sleep(0, loop=loop)
    assert not fut.done()

    yield from pub.execute('publish', 'channel:1', 'hello')
    yield from asyncio.sleep(0, loop=loop)  # read in connection
    yield from asyncio.sleep(0, loop=loop)  # call Future.set_result
    assert fut.done()
    res = yield from fut
    assert res is True


@pytest.mark.run_loop
def test_decode_message(loop):
    mpsc = Receiver(loop)
    ch = mpsc.channel('channel:1')
    ch.put_nowait(b'Some data')

    res = yield from mpsc.get(encoding='utf-8')
    assert isinstance(res[0], _Sender)
    assert res[1] == 'Some data'

    ch.put_nowait('{"hello": "world"}')
    res = yield from mpsc.get(decoder=json.loads)
    assert isinstance(res[0], _Sender)
    assert res[1] == {'hello': 'world'}

    ch.put_nowait(b'{"hello": "world"}')
    res = yield from mpsc.get(encoding='utf-8', decoder=json.loads)
    assert isinstance(res[0], _Sender)
    assert res[1] == {'hello': 'world'}


@pytest.mark.skipif(sys.version_info >= (3, 6),
                    reason="json.loads accept bytes since Python 3.6")
@pytest.mark.run_loop
def test_decode_message_error(loop):
    mpsc = Receiver(loop)
    ch = mpsc.channel('channel:1')

    ch.put_nowait(b'{"hello": "world"}')
    unexpected = (mock.ANY, {'hello': 'world'})
    with pytest.raises(TypeError):
        assert (yield from mpsc.get(decoder=json.loads)) == unexpected

    ch = mpsc.pattern('*')
    ch.put_nowait((b'channel', b'{"hello": "world"}'))
    unexpected = (mock.ANY, b'channel', {'hello': 'world'})
    with pytest.raises(TypeError):
        assert (yield from mpsc.get(decoder=json.loads)) == unexpected


@pytest.mark.run_loop
def test_decode_message_for_pattern(loop):
    mpsc = Receiver(loop)
    ch = mpsc.pattern('*')
    ch.put_nowait((b'channel', b'Some data'))

    res = yield from mpsc.get(encoding='utf-8')
    assert isinstance(res[0], _Sender)
    assert res[1] == (b'channel', 'Some data')

    ch.put_nowait((b'channel', '{"hello": "world"}'))
    res = yield from mpsc.get(decoder=json.loads)
    assert isinstance(res[0], _Sender)
    assert res[1] == (b'channel', {'hello': 'world'})

    ch.put_nowait((b'channel', b'{"hello": "world"}'))
    res = yield from mpsc.get(encoding='utf-8', decoder=json.loads)
    assert isinstance(res[0], _Sender)
    assert res[1] == (b'channel', {'hello': 'world'})



import asyncio
import pytest
import socket
import subprocess
import sys
import contextlib
import os
import ssl
import time
import logging
import tempfile
import atexit

from collections import namedtuple
from async_timeout import timeout as async_timeout

import aioredis
import aioredis.sentinel


TCPAddress = namedtuple('TCPAddress', 'host port')

RedisServer = namedtuple('RedisServer', 'name tcp_address unixsocket version')

SentinelServer = namedtuple('SentinelServer',
                            'name tcp_address unixsocket version masters')

# Public fixtures


@pytest.yield_fixture
def loop():
    """Creates new event loop."""
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(None)

    try:
        yield loop
    finally:
        if hasattr(loop, 'is_closed'):
            closed = loop.is_closed()
        else:
            closed = loop._closed   # XXX
        if not closed:
            loop.call_soon(loop.stop)
            loop.run_forever()
            loop.close()


@pytest.fixture(scope='session')
def unused_port():
    """Gets random free port."""
    def fun():
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(('127.0.0.1', 0))
            return s.getsockname()[1]
    return fun


@pytest.fixture
def create_connection(_closable, loop):
    """Wrapper around aioredis.create_connection."""

    @asyncio.coroutine
    def f(*args, **kw):
        kw.setdefault('loop', loop)
        conn = yield from aioredis.create_connection(*args, **kw)
        _closable(conn)
        return conn
    return f


@pytest.fixture(params=[
    aioredis.create_redis,
    aioredis.create_redis_pool],
    ids=['single', 'pool'])
def create_redis(_closable, loop, request):
    """Wrapper around aioredis.create_redis."""
    factory = request.param

    @asyncio.coroutine
    def f(*args, **kw):
        kw.setdefault('loop', loop)
        redis = yield from factory(*args, **kw)
        _closable(redis)
        return redis
    return f


@pytest.fixture
def create_pool(_closable, loop):
    """Wrapper around aioredis.create_pool."""

    @asyncio.coroutine
    def f(*args, **kw):
        kw.setdefault('loop', loop)
        redis = yield from aioredis.create_pool(*args, **kw)
        _closable(redis)
        return redis
    return f


@pytest.fixture
def create_sentinel(_closable, loop):
    """Helper instantiating RedisSentinel client."""

    @asyncio.coroutine
    def f(*args, **kw):
        kw.setdefault('loop', loop)
        client = yield from aioredis.sentinel.create_sentinel(*args, **kw)
        _closable(client)
        return client
    return f


@pytest.fixture
def pool(create_pool, server, loop):
    """Returns RedisPool instance."""
    pool = loop.run_until_complete(
        create_pool(server.tcp_address, loop=loop))
    return pool


@pytest.fixture
def redis(create_redis, server, loop):
    """Returns Redis client instance."""
    redis = loop.run_until_complete(
        create_redis(server.tcp_address, loop=loop))
    loop.run_until_complete(redis.flushall())
    return redis


@pytest.fixture
def redis_sentinel(create_sentinel, sentinel, loop):
    """Returns Redis Sentinel client instance."""
    redis_sentinel = loop.run_until_complete(
        create_sentinel([sentinel.tcp_address], loop=loop))
    assert loop.run_until_complete(redis_sentinel.ping()) == b'PONG'
    return redis_sentinel


@pytest.yield_fixture
def _closable(loop):
    conns = []

    try:
        yield conns.append
    finally:
        waiters = []
        while conns:
            conn = conns.pop(0)
            conn.close()
            waiters.append(conn.wait_closed())
        if waiters:
            loop.run_until_complete(asyncio.gather(*waiters, loop=loop))


@pytest.fixture(scope='session')
def server(start_server):
    """Starts redis-server instance."""
    return start_server('A')


@pytest.fixture(scope='session')
def serverB(start_server):
    """Starts redis-server instance."""
    return start_server('B')


@pytest.fixture(scope='session')
def sentinel(start_sentinel, request, start_server):
    """Starts redis-sentinel instance with one master -- masterA."""
    # Adding master+slave for normal (no failover) tests:
    master_no_fail = start_server('master-no-fail')
    start_server('slave-no-fail', slaveof=master_no_fail)
    # Adding master+slave for failover test;
    masterA = start_server('masterA')
    start_server('slaveA', slaveof=masterA)
    return start_sentinel('main', masterA, master_no_fail)


# Internal stuff #


def pytest_addoption(parser):
    parser.addoption('--redis-server', default=[],
                     action="append",
                     help="Path to redis-server executable,"
                          " defaults to `%(default)s`")
    parser.addoption('--ssl-cafile', default='tests/ssl/cafile.crt',
                     help="Path to testing SSL CA file")
    parser.addoption('--ssl-dhparam', default='tests/ssl/dhparam.pem',
                     help="Path to testing SSL DH params file")
    parser.addoption('--ssl-cert', default='tests/ssl/cert.pem',
                     help="Path to testing SSL CERT file")
    parser.addoption('--uvloop', default=False,
                     action='store_true',
                     help="Run tests with uvloop")


def _read_server_version(redis_bin):
    args = [redis_bin, '--version']
    with subprocess.Popen(args, stdout=subprocess.PIPE) as proc:
        version = proc.stdout.readline().decode('utf-8')
    for part in version.split():
        if part.startswith('v='):
            break
    else:
        raise RuntimeError(
            "No version info can be found in {}".format(version))
    return tuple(map(int, part[2:].split('.')))


@contextlib.contextmanager
def config_writer(path):
    with open(path, 'wt') as f:
        def write(*args):
            print(*args, file=f)
        yield write


REDIS_SERVERS = []
VERSIONS = {}


def format_version(srv):
    return 'redis_v{}'.format('.'.join(map(str, VERSIONS[srv])))


@pytest.fixture(scope='session', params=REDIS_SERVERS, ids=format_version)
def server_bin(request):
    """Common for start_server and start_sentinel server bin path parameter.
    """
    return request.param


@pytest.fixture(scope='session')
def start_server(_proc, request, unused_port, server_bin):
    """Starts Redis server instance.

    Caches instances by name.
    ``name`` param -- instance alias
    ``config_lines`` -- optional list of config directives to put in config
        (if no config_lines passed -- no config will be generated,
         for backward compatibility).
    """

    version = _read_server_version(server_bin)
    verbose = request.config.getoption('-v') > 3

    servers = {}

    def timeout(t):
        end = time.time() + t
        while time.time() <= end:
            yield True
        raise RuntimeError("Redis startup timeout expired")

    def maker(name, config_lines=None, *, slaveof=None):
        assert slaveof is None or isinstance(slaveof, RedisServer), slaveof
        if name in servers:
            return servers[name]

        port = unused_port()
        tcp_address = TCPAddress('localhost', port)
        if sys.platform == 'win32':
            unixsocket = None
        else:
            unixsocket = '/tmp/aioredis.{}.sock'.format(port)
        dumpfile = 'dump-{}.rdb'.format(port)
        data_dir = tempfile.gettempdir()
        dumpfile_path = os.path.join(data_dir, dumpfile)
        stdout_file = os.path.join(data_dir, 'aioredis.{}.stdout'.format(port))
        tmp_files = [dumpfile_path, stdout_file]
        if config_lines:
            config = os.path.join(data_dir, 'aioredis.{}.conf'.format(port))
            with config_writer(config) as write:
                write('daemonize no')
                write('save ""')
                write('dir ', data_dir)
                write('dbfilename', dumpfile)
                write('port', port)
                if unixsocket:
                    write('unixsocket', unixsocket)
                    tmp_files.append(unixsocket)
                write('# extra config')
                for line in config_lines:
                    write(line)
                if slaveof is not None:
                    write("slaveof {0.tcp_address.host} {0.tcp_address.port}"
                          .format(slaveof))
            args = [config]
            tmp_files.append(config)
        else:
            args = ['--daemonize', 'no',
                    '--save', '""',
                    '--dir', data_dir,
                    '--dbfilename', dumpfile,
                    '--port', str(port),
                    ]
            if unixsocket:
                args += [
                    '--unixsocket', unixsocket,
                    ]
            if slaveof is not None:
                args += [
                    '--slaveof',
                    str(slaveof.tcp_address.host),
                    str(slaveof.tcp_address.port),
                    ]

        f = open(stdout_file, 'w')
        atexit.register(f.close)
        proc = _proc(server_bin, *args,
                     stdout=f,
                     stderr=subprocess.STDOUT,
                     _clear_tmp_files=tmp_files)
        with open(stdout_file, 'rt') as f:
            for _ in timeout(10):
                assert proc.poll() is None, (
                    "Process terminated", proc.returncode)
                log = f.readline()
                if log and verbose:
                    print(name, ":", log, end='')
                if 'The server is now ready to accept connections ' in log:
                    break
            if slaveof is not None:
                for _ in timeout(10):
                    log = f.readline()
                    if log and verbose:
                        print(name, ":", log, end='')
                    if 'sync: Finished with success' in log:
                        break
        info = RedisServer(name, tcp_address, unixsocket, version)
        servers.setdefault(name, info)
        return info

    return maker


@pytest.fixture(scope='session')
def start_sentinel(_proc, request, unused_port, server_bin):
    """Starts Redis Sentinel instances."""
    version = _read_server_version(server_bin)
    verbose = request.config.getoption('-v') > 3

    sentinels = {}

    def timeout(t):
        end = time.time() + t
        while time.time() <= end:
            yield True
        raise RuntimeError("Redis startup timeout expired")

    def maker(name, *masters, quorum=1, noslaves=False):
        key = (name,) + masters
        if key in sentinels:
            return sentinels[key]
        port = unused_port()
        tcp_address = TCPAddress('localhost', port)
        data_dir = tempfile.gettempdir()
        config = os.path.join(
            data_dir, 'aioredis-sentinel.{}.conf'.format(port))
        stdout_file = os.path.join(
            data_dir, 'aioredis-sentinel.{}.stdout'.format(port))
        tmp_files = [config, stdout_file]
        if sys.platform == 'win32':
            unixsocket = None
        else:
            unixsocket = os.path.join(
                data_dir, 'aioredis-sentinel.{}.sock'.format(port))
            tmp_files.append(unixsocket)

        with config_writer(config) as write:
            write('daemonize no')
            write('save ""')
            write('port', port)
            if unixsocket:
                write('unixsocket', unixsocket)
            write('loglevel debug')
            for master in masters:
                write('sentinel monitor', master.name,
                      '127.0.0.1', master.tcp_address.port, quorum)
                write('sentinel down-after-milliseconds', master.name, '3000')
                write('sentinel failover-timeout', master.name, '3000')

        f = open(stdout_file, 'w')
        atexit.register(f.close)
        proc = _proc(server_bin,
                     config,
                     '--sentinel',
                     stdout=f,
                     stderr=subprocess.STDOUT,
                     _clear_tmp_files=tmp_files)
        # XXX: wait sentinel see all masters and slaves;
        all_masters = {m.name for m in masters}
        if noslaves:
            all_slaves = {}
        else:
            all_slaves = {m.name for m in masters}
        with open(stdout_file, 'rt') as f:
            for _ in timeout(30):
                assert proc.poll() is None, (
                    "Process terminated", proc.returncode)
                log = f.readline()
                if log and verbose:
                    print(name, ":", log, end='')
                for m in masters:
                    if '# +monitor master {}'.format(m.name) in log:
                        all_masters.discard(m.name)
                    if '* +slave slave' in log and \
                            '@ {}'.format(m.name) in log:
                        all_slaves.discard(m.name)
                if not all_masters and not all_slaves:
                    break
            else:
                raise RuntimeError("Could not start Sentinel")

        masters = {m.name: m for m in masters}
        info = SentinelServer(name, tcp_address, unixsocket, version, masters)
        sentinels.setdefault(key, info)
        return info
    return maker


@pytest.fixture(scope='session')
def ssl_proxy(_proc, request, unused_port):
    by_port = {}

    cafile = os.path.abspath(request.config.getoption('--ssl-cafile'))
    certfile = os.path.abspath(request.config.getoption('--ssl-cert'))
    dhfile = os.path.abspath(request.config.getoption('--ssl-dhparam'))
    assert os.path.exists(cafile), \
        "Missing SSL CA file, run `make certificate` to generate new one"
    assert os.path.exists(certfile), \
        "Missing SSL CERT file, run `make certificate` to generate new one"
    assert os.path.exists(dhfile), \
        "Missing SSL DH params, run `make certificate` to generate new one"

    if hasattr(ssl, 'create_default_context'):
        ssl_ctx = ssl.create_default_context(cafile=cafile)
    else:
        ssl_ctx = ssl.SSLContext(ssl.PROTOCOL_SSLv23)
        ssl_ctx.load_verify_locations(cafile=cafile)
    if hasattr(ssl_ctx, 'check_hostname'):
        # available since python 3.4
        ssl_ctx.check_hostname = False
    ssl_ctx.verify_mode = ssl.CERT_NONE
    ssl_ctx.load_dh_params(dhfile)

    def sockat(unsecure_port):
        if unsecure_port in by_port:
            return by_port[unsecure_port]

        secure_port = unused_port()
        _proc('/usr/bin/socat',
              'openssl-listen:{port},'
              'dhparam={param},'
              'cert={cert},verify=0,fork'
              .format(port=secure_port, param=dhfile, cert=certfile),
              'tcp-connect:localhost:{}'
              .format(unsecure_port)
              )
        time.sleep(1)   # XXX
        by_port[unsecure_port] = secure_port, ssl_ctx
        return secure_port, ssl_ctx

    return sockat


@pytest.yield_fixture(scope='session')
def _proc():
    processes = []
    tmp_files = set()

    def run(*commandline, _clear_tmp_files=(), **kwargs):
        proc = subprocess.Popen(commandline, **kwargs)
        processes.append(proc)
        tmp_files.update(_clear_tmp_files)
        return proc

    try:
        yield run
    finally:
        while processes:
            proc = processes.pop(0)
            proc.terminate()
            proc.wait()
        for path in tmp_files:
            try:
                os.remove(path)
            except FileNotFoundError:
                pass


@pytest.mark.tryfirst
def pytest_pycollect_makeitem(collector, name, obj):
    if collector.funcnamefilter(name):
        if not callable(obj):
            return
        item = pytest.Function(name, parent=collector)
        if 'run_loop' in item.keywords:
            # TODO: re-wrap with asyncio.coroutine if not native coroutine
            return list(collector._genfunctions(name, obj))


@pytest.mark.tryfirst
def pytest_pyfunc_call(pyfuncitem):
    """
    Run asyncio marked test functions in an event loop instead of a normal
    function call.
    """
    if 'run_loop' in pyfuncitem.keywords:
        marker = pyfuncitem.keywords['run_loop']
        funcargs = pyfuncitem.funcargs
        loop = funcargs['loop']
        testargs = {arg: funcargs[arg]
                    for arg in pyfuncitem._fixtureinfo.argnames}

        loop.run_until_complete(
            _wait_coro(pyfuncitem.obj, testargs,
                       timeout=marker.kwargs.get('timeout', 15),
                       loop=loop))
        return True


@asyncio.coroutine
def _wait_coro(corofunc, kwargs, timeout, loop):
    with async_timeout(timeout, loop=loop):
        return (yield from corofunc(**kwargs))


def pytest_runtest_setup(item):
    if 'run_loop' in item.keywords and 'loop' not in item.fixturenames:
        # inject an event loop fixture for all async tests
        item.fixturenames.append('loop')


def pytest_ignore_collect(path, config):
    if 'py35' in str(path):
        if sys.version_info < (3, 5, 0):
            return True


def pytest_collection_modifyitems(session, config, items):
    for item in items:
        if 'redis_version' in item.keywords:
            marker = item.keywords['redis_version']
            try:
                version = VERSIONS[item.callspec.getparam('server_bin')]
            except (KeyError, ValueError, AttributeError):
                # TODO: throw noisy warning
                continue
            if version < marker.kwargs['version']:
                item.add_marker(pytest.mark.skip(
                    reason=marker.kwargs['reason']))
        if 'ssl_proxy' in item.fixturenames:
            item.add_marker(pytest.mark.skipif(
                "not os.path.exists('/usr/bin/socat')",
                reason="socat package required (apt-get install socat)"))


def pytest_configure(config):
    bins = config.getoption('--redis-server')[:]
    REDIS_SERVERS[:] = bins or ['/usr/bin/redis-server']
    VERSIONS.update({srv: _read_server_version(srv)
                     for srv in REDIS_SERVERS})
    assert VERSIONS, ("Expected to detect redis versions", REDIS_SERVERS)
    if config.getoption('--uvloop'):
        try:
            import uvloop
        except ImportError:
            raise RuntimeError(
                "Can not import uvloop, make sure it is installed")
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


def logs(logger, level=None):
    """Catches logs for given logger and level.

    See unittest.TestCase.assertLogs for details.
    """
    return _AssertLogsContext(logger, level)


_LoggingWatcher = namedtuple("_LoggingWatcher", ["records", "output"])


class _CapturingHandler(logging.Handler):
    """
    A logging handler capturing all (raw and formatted) logging output.
    """

    def __init__(self):
        logging.Handler.__init__(self)
        self.watcher = _LoggingWatcher([], [])

    def flush(self):
        pass

    def emit(self, record):
        self.watcher.records.append(record)
        msg = self.format(record)
        self.watcher.output.append(msg)


class _AssertLogsContext:
    """Standard unittest's _AssertLogsContext context manager
    adopted to raise pytest failure.
    """
    LOGGING_FORMAT = "%(levelname)s:%(name)s:%(message)s"

    def __init__(self, logger_name, level):
        self.logger_name = logger_name
        if level:
            self.level = level
        else:
            self.level = logging.INFO
        self.msg = None

    def __enter__(self):
        if isinstance(self.logger_name, logging.Logger):
            logger = self.logger = self.logger_name
        else:
            logger = self.logger = logging.getLogger(self.logger_name)
        formatter = logging.Formatter(self.LOGGING_FORMAT)
        handler = _CapturingHandler()
        handler.setFormatter(formatter)
        self.watcher = handler.watcher
        self.old_handlers = logger.handlers[:]
        self.old_level = logger.level
        self.old_propagate = logger.propagate
        logger.handlers = [handler]
        logger.setLevel(self.level)
        logger.propagate = False
        return handler.watcher

    def __exit__(self, exc_type, exc_value, tb):
        self.logger.handlers = self.old_handlers
        self.logger.propagate = self.old_propagate
        self.logger.setLevel(self.old_level)
        if exc_type is not None:
            # let unexpected exceptions pass through
            return False
        if len(self.watcher.records) == 0:
            pytest.fail(
                "no logs of level {} or higher triggered on {}"
                .format(logging.getLevelName(self.level), self.logger.name))


def redis_version(*version, reason):
    assert 1 < len(version) <= 3, version
    assert all(isinstance(v, int) for v in version), version
    return pytest.mark.redis_version(version=version, reason=reason)


def assert_almost_equal(first, second, places=None, msg=None, delta=None):
    assert not (places is None and delta is None), \
        "Both places and delta are not set, please set one"
    if delta is not None:
        assert abs(first - second) <= delta
    else:
        assert round(abs(first - second), places) == 0


def pytest_namespace():
    return {
        'assert_almost_equal': assert_almost_equal,
        'redis_version': redis_version,
        'logs': logs,
        }



import asyncio
import pytest


@asyncio.coroutine
def add(redis, key, members):
    ok = yield from redis.connection.execute(b'sadd', key, members)
    assert ok == 1


@pytest.mark.run_loop
def test_sadd(redis):
    key, member = b'key:sadd', b'hello'
    # add member to the set, expected result: 1
    test_result = yield from redis.sadd(key, member)
    assert test_result == 1

    # add other value, expected result: 1
    test_result = yield from redis.sadd(key, b'world')
    assert test_result == 1

    # add existing member to the set, expected result: 0
    test_result = yield from redis.sadd(key, member)
    assert test_result == 0

    with pytest.raises(TypeError):
        yield from redis.sadd(None, 10)


@pytest.mark.run_loop
def test_scard(redis):
    key, member = b'key:scard', b'hello'

    # check that our set is empty one
    empty_size = yield from redis.scard(key)
    assert empty_size == 0

    # add more members to the set and check, set size on every step
    for i in range(1, 11):
        incr = str(i).encode('utf-8')
        yield from add(redis, key, member + incr)
        current_size = yield from redis.scard(key)
        assert current_size == i

    with pytest.raises(TypeError):
        yield from redis.scard(None)


@pytest.mark.run_loop
def test_sdiff(redis):
    key1 = b'key:sdiff:1'
    key2 = b'key:sdiff:2'
    key3 = b'key:sdiff:3'

    members1 = (b'a', b'b', b'c', b'd')
    members2 = (b'c',)
    members3 = (b'a', b'c', b'e')

    yield from redis.sadd(key1, *members1)
    yield from redis.sadd(key2, *members2)
    yield from redis.sadd(key3, *members3)

    # test multiple keys
    test_result = yield from redis.sdiff(key1, key2, key3)
    assert set(test_result) == {b'b', b'd'}

    # test single key
    test_result = yield from redis.sdiff(key2)
    assert set(test_result) == {b'c'}

    with pytest.raises(TypeError):
        yield from redis.sdiff(None)
    with pytest.raises(TypeError):
        yield from redis.sdiff(key1, None)


@pytest.mark.run_loop
def test_sdiffstore(redis):
    key1 = b'key:sdiffstore:1'
    key2 = b'key:sdiffstore:2'
    destkey = b'key:sdiffstore:destkey'
    members1 = (b'a', b'b', b'c')
    members2 = (b'c', b'd', b'e')

    yield from redis.sadd(key1, *members1)
    yield from redis.sadd(key2, *members2)

    # test basic use case, expected: since diff contains only two members
    test_result = yield from redis.sdiffstore(destkey, key1, key2)
    assert test_result == 2

    # make sure that destkey contains 2 members
    test_result = yield from redis.scard(destkey)
    assert test_result == 2

    # try sdiffstore in case none of sets exists
    test_result = yield from redis.sdiffstore(
        b'not:' + destkey, b'not:' + key1, b'not:' + key2)
    assert test_result == 0

    with pytest.raises(TypeError):
        yield from redis.sdiffstore(None, key1)
    with pytest.raises(TypeError):
        yield from redis.sdiffstore(destkey, None)
    with pytest.raises(TypeError):
        yield from redis.sdiffstore(destkey, key1, None)


@pytest.mark.run_loop
def test_sinter(redis):
    key1 = b'key:sinter:1'
    key2 = b'key:sinter:2'
    key3 = b'key:sinter:3'

    members1 = (b'a', b'b', b'c', b'd')
    members2 = (b'c',)
    members3 = (b'a', b'c', b'e')

    yield from redis.sadd(key1, *members1)
    yield from redis.sadd(key2, *members2)
    yield from redis.sadd(key3, *members3)

    # test multiple keys
    test_result = yield from redis.sinter(key1, key2, key3)
    assert set(test_result) == {b'c'}

    # test single key
    test_result = yield from redis.sinter(key2)
    assert set(test_result) == {b'c'}

    with pytest.raises(TypeError):
        yield from redis.sinter(None)
    with pytest.raises(TypeError):
        yield from redis.sinter(key1, None)


@pytest.mark.run_loop
def test_sinterstore(redis):
    key1 = b'key:sinterstore:1'
    key2 = b'key:sinterstore:2'
    destkey = b'key:sinterstore:destkey'
    members1 = (b'a', b'b', b'c')
    members2 = (b'c', b'd', b'e')

    yield from redis.sadd(key1, *members1)
    yield from redis.sadd(key2, *members2)

    # test basic use case, expected: since inter contains only one member
    test_result = yield from redis.sinterstore(destkey, key1, key2)
    assert test_result == 1

    # make sure that destkey contains only one member
    test_result = yield from redis.scard(destkey)
    assert test_result == 1

    # try sinterstore in case none of sets exists
    test_result = yield from redis.sinterstore(
        b'not:' + destkey, b'not:' + key1, b'not:' + key2)
    assert test_result == 0

    with pytest.raises(TypeError):
        yield from redis.sinterstore(None, key1)
    with pytest.raises(TypeError):
        yield from redis.sinterstore(destkey, None)
    with pytest.raises(TypeError):
        yield from redis.sinterstore(destkey, key1, None)


@pytest.mark.run_loop
def test_sismember(redis):
    key, member = b'key:sismember', b'hello'
    # add member to the set, expected result: 1
    test_result = yield from redis.sadd(key, member)
    assert test_result == 1

    # test that value in set
    test_result = yield from redis.sismember(key, member)
    assert test_result == 1
    # test that value not in set
    test_result = yield from redis.sismember(key, b'world')
    assert test_result == 0

    with pytest.raises(TypeError):
        yield from redis.sismember(None, b'world')


@pytest.mark.run_loop
def test_smembers(redis):
    key = b'key:smembers'
    member1 = b'hello'
    member2 = b'world'

    yield from redis.sadd(key, member1)
    yield from redis.sadd(key, member2)

    # test not empty set
    test_result = yield from redis.smembers(key)
    assert set(test_result) == {member1, member2}

    # test empty set
    test_result = yield from redis.smembers(b'not:' + key)
    assert test_result == []

    # test encoding param
    test_result = yield from redis.smembers(key, encoding='utf-8')
    assert set(test_result) == {'hello', 'world'}

    with pytest.raises(TypeError):
        yield from redis.smembers(None)


@pytest.mark.run_loop
def test_smove(redis):
    key1 = b'key:smove:1'
    key2 = b'key:smove:2'
    member1 = b'one'
    member2 = b'two'
    member3 = b'three'
    yield from redis.sadd(key1, member1, member2)
    yield from redis.sadd(key2, member3)
    # move member2 to second set
    test_result = yield from redis.smove(key1, key2, member2)
    assert test_result == 1
    # check first set, member should be removed
    test_result = yield from redis.smembers(key1)
    assert test_result == [member1]
    # check second set, member should be added
    test_result = yield from redis.smembers(key2)
    assert set(test_result) == {member2, member3}

    # move to empty set
    test_result = yield from redis.smove(
        key1, b'not:' + key2, member1)
    assert test_result == 1

    # move from empty set (set with under key1 is empty now
    test_result = yield from redis.smove(
        key1, b'not:' + key2, member1)
    assert test_result == 0

    # move from set that does not exists to set tha does not exists too
    test_result = yield from redis.smove(
        b'not:' + key1, b'other:not:' + key2, member1)
    assert test_result == 0

    with pytest.raises(TypeError):
        yield from redis.smove(None, key1, member1)
    with pytest.raises(TypeError):
        yield from redis.smove(key1, None, member1)


@pytest.mark.run_loop
def test_spop(redis):
    key = b'key:spop:1'
    members = b'one', b'two', b'three'
    yield from redis.sadd(key, *members)

    for _ in members:
        test_result = yield from redis.spop(key)
        assert test_result in members

    # test with encoding
    members = 'four', 'five', 'six'
    yield from redis.sadd(key, *members)

    for _ in members:
        test_result = yield from redis.spop(key, encoding='utf-8')
        assert test_result in members

    # make sure set is empty, after all values poped
    test_result = yield from redis.smembers(key)
    assert test_result == []

    # try to pop data from empty set
    test_result = yield from redis.spop(b'not:' + key)
    assert test_result is None

    with pytest.raises(TypeError):
        yield from redis.spop(None)


@pytest.mark.run_loop
def test_srandmember(redis):
    key = b'key:srandmember:1'
    members = b'one', b'two', b'three', b'four', b'five', b'six', b'seven'
    yield from redis.sadd(key, *members)

    for _ in members:
        test_result = yield from redis.srandmember(key)
        assert test_result in members

    # test with encoding
    test_result = yield from redis.srandmember(key, encoding='utf-8')
    strings = {'one', 'two', 'three', 'four', 'five', 'six', 'seven'}
    assert test_result in strings

    # make sure set contains all values, and nothing missing
    test_result = yield from redis.smembers(key)
    assert set(test_result) == set(members)

    # fetch 4 elements for the first time, as result 4 distinct values
    test_result1 = yield from redis.srandmember(key, 4)
    assert len(test_result1) == 4
    assert set(test_result1).issubset(members) is True

    # test negative count, same element may be returned multiple times
    test_result2 = yield from redis.srandmember(key, -10)
    assert len(test_result2) == 10
    assert set(test_result2).issubset(members) is True
    assert len(set(test_result2)) <= len(members)

    # pull member from empty set
    test_result = yield from redis.srandmember(b'not' + key)
    assert test_result is None

    with pytest.raises(TypeError):
        yield from redis.srandmember(None)


@pytest.mark.run_loop
def test_srem(redis):
    key = b'key:srem:1'
    members = b'one', b'two', b'three', b'four', b'five', b'six', b'seven'
    yield from redis.sadd(key, *members)

    # remove one element from set
    test_result = yield from redis.srem(key, members[-1])
    assert test_result == 1

    # remove not existing element
    test_result = yield from redis.srem(key, b'foo')
    assert test_result == 0

    # remove not existing element from not existing set
    test_result = yield from redis.srem(b'not:' + key, b'foo')
    assert test_result == 0

    # remove multiple elements from set
    test_result = yield from redis.srem(key, *members[:-1])
    assert test_result == 6
    with pytest.raises(TypeError):
        yield from redis.srem(None, members)


@pytest.mark.run_loop
def test_sunion(redis):
    key1 = b'key:sunion:1'
    key2 = b'key:sunion:2'
    key3 = b'key:sunion:3'

    members1 = [b'a', b'b', b'c', b'd']
    members2 = [b'c']
    members3 = [b'a', b'c', b'e']

    yield from redis.sadd(key1, *members1)
    yield from redis.sadd(key2, *members2)
    yield from redis.sadd(key3, *members3)

    # test multiple keys
    test_result = yield from redis.sunion(key1, key2, key3)
    assert set(test_result) == set(members1 + members2 + members3)

    # test single key
    test_result = yield from redis.sunion(key2)
    assert set(test_result) == {b'c'}

    with pytest.raises(TypeError):
        yield from redis.sunion(None)
    with pytest.raises(TypeError):
        yield from redis.sunion(key1, None)


@pytest.mark.run_loop
def test_sunionstore(redis):
    key1 = b'key:sunionstore:1'
    key2 = b'key:sunionstore:2'
    destkey = b'key:sunionstore:destkey'
    members1 = (b'a', b'b', b'c')
    members2 = (b'c', b'd', b'e')

    yield from redis.sadd(key1, *members1)
    yield from redis.sadd(key2, *members2)

    # test basic use case
    test_result = yield from redis.sunionstore(destkey, key1, key2)
    assert test_result == 5

    # make sure that destkey contains 5 members
    test_result = yield from redis.scard(destkey)
    assert test_result == 5

    # try sunionstore in case none of sets exists
    test_result = yield from redis.sunionstore(
        b'not:' + destkey, b'not:' + key1, b'not:' + key2)
    assert test_result == 0

    with pytest.raises(TypeError):
        yield from redis.sunionstore(None, key1)
    with pytest.raises(TypeError):
        yield from redis.sunionstore(destkey, None)
    with pytest.raises(TypeError):
        yield from redis.sunionstore(destkey, key1, None)


@pytest.redis_version(2, 8, 0, reason='SSCAN is available since redis>=2.8.0')
@pytest.mark.run_loop
def test_sscan(redis):
    key = b'key:sscan'
    for i in range(1, 11):
        foo_or_bar = 'bar' if i % 3 else 'foo'
        member = 'member:{}:{}'.format(foo_or_bar, i).encode('utf-8')
        yield from add(redis, key, member)

    cursor, values = yield from redis.sscan(
        key, match=b'member:foo:*')
    assert len(values) == 3

    cursor, values = yield from redis.sscan(
        key, match=b'member:bar:*')
    assert len(values) == 7

    # SCAN family functions do not guarantee that the number (count) of
    # elements returned per call are in a given range. So here
    # just dummy test, that *count* argument does not break something
    cursor = b'0'
    test_values = []
    while cursor:
        cursor, values = yield from redis.sscan(key, cursor, count=2)
        test_values.extend(values)
    assert len(test_values) == 10

    with pytest.raises(TypeError):
        yield from redis.sscan(None)



import asyncio
import aioredis


async def main():
    pool = await aioredis.create_pool(
        ('localhost', 6379),
        minsize=5, maxsize=10)
    with await pool as conn:    # low-level redis connection
        await conn.execute('set', 'my-key', 'value')
        val = await conn.execute('get', 'my-key')
    print('raw value:', val)
    pool.close()
    await pool.wait_closed()    # closing all open connections


if __name__ == '__main__':
    asyncio.get_event_loop().run_until_complete(main())



import asyncio
import aioredis


@asyncio.coroutine
def main():
    redis = yield from aioredis.create_redis(
        ('localhost', 6379))

    # No pipelining;
    @asyncio.coroutine
    def wait_each_command():
        val = yield from redis.get('foo')    # wait until `val` is available
        cnt = yield from redis.incr('bar')   # wait until `cnt` is available
        return val, cnt

    # Sending multiple commands and then gathering results
    @asyncio.coroutine
    def pipelined():
        fut1 = redis.get('foo')      # issue command and return future
        fut2 = redis.incr('bar')     # issue command and return future
        # block until results are available
        val, cnt = yield from asyncio.gather(fut1, fut2)
        return val, cnt

    # Convenient way
    @asyncio.coroutine
    def convenience_way():
        pipe = redis.pipeline()
        fut1 = pipe.get('foo')
        fut2 = pipe.incr('bar')
        result = yield from pipe.execute()
        val, cnt = yield from asyncio.gather(fut1, fut2)
        assert result == [val, cnt]
        return val, cnt

    res = yield from wait_each_command()
    print(res)
    res = yield from pipelined()
    print(res)
    res = yield from convenience_way()
    print(res)

    redis.close()
    yield from redis.wait_closed()


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



import asyncio
import aioredis


def main():
    """Scan command example."""
    loop = asyncio.get_event_loop()

    @asyncio.coroutine
    def go():
        redis = yield from aioredis.create_redis(
            ('localhost', 6379))

        yield from redis.mset('key:1', 'value1', 'key:2', 'value2')
        cur = b'0'  # set initial cursor to 0
        while cur:
            cur, keys = yield from redis.scan(cur, match='key:*')
            print("Iteration results:", keys)

        redis.close()
        yield from redis.wait_closed()
    loop.run_until_complete(go())


if __name__ == '__main__':
    import os
    if 'redis_version:2.6' not in os.environ.get('REDIS_VERSION', ''):
        main()



import pytest


@pytest.redis_version(2, 8, 0, reason='SSCAN is available since redis>=2.8.0')
@pytest.mark.run_loop
async def test_isscan(redis):
    key = b'key:sscan'
    for i in range(1, 11):
        foo_or_bar = 'bar' if i % 3 else 'foo'
        member = 'member:{}:{}'.format(foo_or_bar, i).encode('utf-8')
        assert await redis.sadd(key, member) == 1

    async def coro(cmd):
        lst = []
        async for i in cmd:
            lst.append(i)
        return lst

    ret = await coro(redis.isscan(key, match=b'member:foo:*'))
    assert set(ret) == {b'member:foo:3',
                        b'member:foo:6',
                        b'member:foo:9'}

    ret = await coro(redis.isscan(key, match=b'member:bar:*'))
    assert set(ret) == {b'member:bar:1',
                        b'member:bar:2',
                        b'member:bar:4',
                        b'member:bar:5',
                        b'member:bar:7',
                        b'member:bar:8',
                        b'member:bar:10'}

    # SCAN family functions do not guarantee that the number (count) of
    # elements returned per call are in a given range. So here
    # just dummy test, that *count* argument does not break something
    ret = await coro(redis.isscan(key, count=2))
    assert set(ret) == {b'member:foo:3',
                        b'member:foo:6',
                        b'member:foo:9',
                        b'member:bar:1',
                        b'member:bar:2',
                        b'member:bar:4',
                        b'member:bar:5',
                        b'member:bar:7',
                        b'member:bar:8',
                        b'member:bar:10'}

    with pytest.raises(TypeError):
        await redis.isscan(None)



import asyncio
import aioredis


async def main():
    redis = await aioredis.create_redis(
        ('localhost', 6379))
    await redis.delete('foo', 'bar')
    tr = redis.multi_exec()
    fut1 = tr.incr('foo')
    fut2 = tr.incr('bar')
    res = await tr.execute()
    res2 = await asyncio.gather(fut1, fut2)
    print(res)
    assert res == res2

    redis.close()
    await redis.wait_closed()


if __name__ == '__main__':
    asyncio.get_event_loop().run_until_complete(main())



import asyncio
import aioredis


@asyncio.coroutine
def main():
    redis = yield from aioredis.create_redis(
        ('localhost', 6379))

    @asyncio.coroutine
    def transaction():
        tr = redis.multi_exec()
        future1 = tr.set('foo', '123')
        future2 = tr.set('bar', '321')
        result = yield from tr.execute()
        assert result == (yield from asyncio.gather(future1, future2))
        return result

    redis.close()
    yield from redis.wait_closed()


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



import pytest
import asyncio
import sys

from aioredis import (
    SlaveNotFoundError,
    ReadOnlyError,
    )


pytestmark = pytest.redis_version(2, 8, 12, reason="Sentinel v2 required")
if sys.platform == 'win32':
    pytestmark = pytest.mark.skip(reason="unstable on windows")


@pytest.mark.xfail
@pytest.mark.run_loop(timeout=40)
def test_auto_failover(start_sentinel, start_server,
                       create_sentinel, create_connection, loop):
    server1 = start_server('master-failover', ['slave-read-only yes'])
    start_server('slave-failover1', ['slave-read-only yes'], slaveof=server1)
    start_server('slave-failover2', ['slave-read-only yes'], slaveof=server1)

    sentinel1 = start_sentinel('sentinel-failover1', server1, quorum=2)
    sentinel2 = start_sentinel('sentinel-failover2', server1, quorum=2)

    sp = yield from create_sentinel([sentinel1.tcp_address,
                                     sentinel2.tcp_address])

    _, old_port = yield from sp.master_address(server1.name)
    # ignoring host
    assert old_port == server1.tcp_address.port
    master = sp.master_for(server1.name)
    res = yield from master.role()
    assert res.role == 'master'
    assert master.address is not None
    assert master.address[1] == old_port

    # wait failover
    conn = yield from create_connection(server1.tcp_address)
    yield from conn.execute("debug", "sleep", 6)
    yield from asyncio.sleep(3, loop=loop)

    # _, new_port = yield from sp.master_address(server1.name)
    # assert new_port != old_port
    # assert new_port == server2.tcp_address.port
    assert (yield from master.set("key", "val"))
    assert master.address is not None
    assert master.address[1] != old_port


@pytest.mark.run_loop
def test_sentinel_normal(sentinel, create_sentinel):
    redis_sentinel = yield from create_sentinel([sentinel.tcp_address])
    redis = redis_sentinel.master_for('masterA')

    info = yield from redis.role()
    assert info.role == 'master'

    key, field, value = b'key:hset', b'bar', b'zap'
    exists = yield from redis.hexists(key, field)
    if exists:
        ret = yield from redis.hdel(key, field)
        assert ret != 1

    ret = yield from redis.hset(key, field, value)
    assert ret == 1
    ret = yield from redis.hset(key, field, value)
    assert ret == 0


@pytest.mark.xfail(reason="same sentinel; single master;")
@pytest.mark.run_loop
def test_sentinel_slave(sentinel, create_sentinel):
    redis_sentinel = yield from create_sentinel([sentinel.tcp_address])
    redis = redis_sentinel.slave_for('masterA')

    info = yield from redis.role()
    assert info.role == 'slave'

    key, field, value = b'key:hset', b'bar', b'zap'
    # redis = yield from get_slave_connection()
    exists = yield from redis.hexists(key, field)
    if exists:
        with pytest.raises(ReadOnlyError):
            yield from redis.hdel(key, field)

    with pytest.raises(ReadOnlyError):
        yield from redis.hset(key, field, value)


@pytest.mark.xfail(reason="Need proper sentinel configuration")
@pytest.mark.run_loop       # (timeout=600)
def test_sentinel_slave_fail(sentinel, create_sentinel, loop):
    redis_sentinel = yield from create_sentinel([sentinel.tcp_address])

    key, field, value = b'key:hset', b'bar', b'zap'

    redis = redis_sentinel.slave_for('masterA')
    exists = yield from redis.hexists(key, field)
    if exists:
        with pytest.raises(ReadOnlyError):
            yield from redis.hdel(key, field)

    with pytest.raises(ReadOnlyError):
        yield from redis.hset(key, field, value)

    ret = yield from redis_sentinel.failover('masterA')
    assert ret is True
    yield from asyncio.sleep(2, loop=loop)

    with pytest.raises(ReadOnlyError):
        yield from redis.hset(key, field, value)

    ret = yield from redis_sentinel.failover('masterA')
    assert ret is True
    yield from asyncio.sleep(2, loop=loop)
    while True:
        try:
            yield from asyncio.sleep(1, loop=loop)
            yield from redis.hset(key, field, value)
        except SlaveNotFoundError:
            continue
        except ReadOnlyError:
            break


@pytest.mark.xfail(reason="Need proper sentinel configuration")
@pytest.mark.run_loop
def test_sentinel_normal_fail(sentinel, create_sentinel, loop):
    redis_sentinel = yield from create_sentinel([sentinel.tcp_address])

    key, field, value = b'key:hset', b'bar', b'zap'
    redis = redis_sentinel.master_for('masterA')
    exists = yield from redis.hexists(key, field)
    if exists:
        ret = yield from redis.hdel(key, field)
        assert ret == 1

    ret = yield from redis.hset(key, field, value)
    assert ret == 1
    ret = yield from redis_sentinel.failover('masterA')
    assert ret is True
    yield from asyncio.sleep(2, loop=loop)
    ret = yield from redis.hset(key, field, value)
    assert ret == 0
    ret = yield from redis_sentinel.failover('masterA')
    assert ret is True
    yield from asyncio.sleep(2, loop=loop)
    redis = redis_sentinel.slave_for('masterA')
    while True:
        try:
            yield from redis.hset(key, field, value)
            yield from asyncio.sleep(1, loop=loop)
            # redis = yield from get_slave_connection()
        except ReadOnlyError:
            break


@pytest.mark.xfail(reason="same sentinel; single master;")
@pytest.mark.run_loop
def test_failover_command(sentinel, create_sentinel, loop):
    master_name = 'masterA'
    redis_sentinel = yield from create_sentinel([sentinel.tcp_address])

    orig_master = yield from redis_sentinel.master_address(master_name)
    ret = yield from redis_sentinel.failover(master_name)
    assert ret is True
    yield from asyncio.sleep(2, loop=loop)

    new_master = yield from redis_sentinel.master_address(master_name)
    assert orig_master != new_master

    ret = yield from redis_sentinel.failover(master_name)
    assert ret is True
    yield from asyncio.sleep(2, loop=loop)

    new_master = yield from redis_sentinel.master_address(master_name)
    assert orig_master == new_master

    redis = redis_sentinel.slave_for(master_name)
    key, field, value = b'key:hset', b'bar', b'zap'
    while True:
        try:
            yield from asyncio.sleep(1, loop=loop)
            yield from redis.hset(key, field, value)
        except SlaveNotFoundError:
            pass
        except ReadOnlyError:
            break



from collections import namedtuple

from aioredis.util import wait_convert, _NOTSET


GeoPoint = namedtuple('GeoPoint', ('longitude', 'latitude'))
GeoMember = namedtuple('GeoMember', ('member', 'dist', 'hash', 'coord'))


class GeoCommandsMixin:
    """Geo commands mixin.

    For commands details see: http://redis.io/commands#geo
    """

    def geoadd(self, key, longitude, latitude, member, *args, **kwargs):
        """Add one or more geospatial items in the geospatial index represented
        using a sorted set.
        """
        return self.execute(
            b'GEOADD', key, longitude, latitude, member, *args, **kwargs
        )

    def geohash(self, key, member, *args, **kwargs):
        """Returns members of a geospatial index as standard geohash strings.
        """
        return self.execute(
            b'GEOHASH', key, member, *args, **kwargs
        )

    def geopos(self, key, member, *args, **kwargs):
        """Returns longitude and latitude of members of a geospatial index.

        :rtype: list[GeoPoint or None]
        """
        fut = self.execute(b'GEOPOS', key, member, *args, **kwargs)
        return wait_convert(fut, make_geopos)

    def geodist(self, key, member1, member2, unit='m'):
        """Returns the distance between two members of a geospatial index.

        :rtype: list[float or None]
        """
        fut = self.execute(b'GEODIST', key, member1, member2, unit)
        return wait_convert(fut, make_geodist)

    def georadius(self, key, longitude, latitude, radius, unit='m', *,
                  with_dist=False, with_hash=False, with_coord=False,
                  count=None, sort=None, encoding=_NOTSET):
        """Query a sorted set representing a geospatial index to fetch members
        matching a given maximum distance from a point.

        Return value follows Redis convention:

        * if none of ``WITH*`` flags are set -- list of strings returned:

            >>> await redis.georadius('Sicily', 15, 37, 200, 'km')
            [b"Palermo", b"Catania"]

        * if any flag (or all) is set -- list of named tuples returned:

            >>> await redis.georadius('Sicily', 15, 37, 200, 'km',
            ...                       with_dist=True)
            [GeoMember(name=b"Palermo", dist=190.4424, hash=None, coord=None),
             GeoMember(name=b"Catania", dist=56.4413, hash=None, coord=None)]

        :raises TypeError: radius is not float or int
        :raises TypeError: count is not int
        :raises ValueError: if unit not equal ``m``, ``km``, ``mi`` or ``ft``
        :raises ValueError: if sort not equal ``ASC`` or ``DESC``

        :rtype: list[str] or list[GeoMember]
        """
        args = validate_georadius_options(
            radius, unit, with_dist, with_hash, with_coord, count, sort
        )

        fut = self.execute(
            b'GEORADIUS', key, longitude, latitude, radius,
            unit, *args, encoding=encoding
        )
        if with_dist or with_hash or with_coord:
            return wait_convert(fut, make_geomember,
                                with_dist=with_dist,
                                with_hash=with_hash,
                                with_coord=with_coord)
        return fut

    def georadiusbymember(self, key, member, radius, unit='m', *,
                          with_dist=False, with_hash=False, with_coord=False,
                          count=None, sort=None, encoding=_NOTSET):
        """Query a sorted set representing a geospatial index to fetch members
        matching a given maximum distance from a member.

        Return value follows Redis convention:

        * if none of ``WITH*`` flags are set -- list of strings returned:

            >>> await redis.georadiusbymember('Sicily', 'Palermo', 200, 'km')
            [b"Palermo", b"Catania"]

        * if any flag (or all) is set -- list of named tuples returned:

            >>> await redis.georadiusbymember('Sicily', 'Palermo', 200, 'km',
            ...                               with_dist=True)
            [GeoMember(name=b"Palermo", dist=190.4424, hash=None, coord=None),
             GeoMember(name=b"Catania", dist=56.4413, hash=None, coord=None)]

        :raises TypeError: radius is not float or int
        :raises TypeError: count is not int
        :raises ValueError: if unit not equal ``m``, ``km``, ``mi`` or ``ft``
        :raises ValueError: if sort not equal ``ASC`` or ``DESC``

        :rtype: list[str] or list[GeoMember]
        """
        args = validate_georadius_options(
            radius, unit, with_dist, with_hash, with_coord, count, sort
        )

        fut = self.execute(
            b'GEORADIUSBYMEMBER', key, member, radius,
            unit, *args, encoding=encoding)
        if with_dist or with_hash or with_coord:
            return wait_convert(fut, make_geomember,
                                with_dist=with_dist,
                                with_hash=with_hash,
                                with_coord=with_coord)
        return fut


def validate_georadius_options(radius, unit, with_dist, with_hash, with_coord,
                               count, sort):
    args = []

    if with_dist:
        args.append(b'WITHDIST')
    if with_hash:
        args.append(b'WITHHASH')
    if with_coord:
        args.append(b'WITHCOORD')

    if unit not in ['m', 'km', 'mi', 'ft']:
        raise ValueError("unit argument must be 'm', 'km', 'mi' or 'ft'")
    if not isinstance(radius, (int, float)):
        raise TypeError("radius argument must be int or float")
    if count:
        if not isinstance(count, int):
            raise TypeError("count argument must be int")
        args += [b'COUNT', count]
    if sort:
        if sort not in ['ASC', 'DESC']:
            raise ValueError("sort argument must be euqal 'ASC' or 'DESC'")
        args.append(sort)
    return args


def make_geocoord(value):
    if isinstance(value, list):
        return GeoPoint(*map(float, value))
    return value


def make_geodist(value):
    if value:
        return float(value)
    return value


def make_geopos(value):
    return [make_geocoord(val) for val in value]


def make_geomember(value, with_dist, with_coord, with_hash):
    res_rows = []

    for row in value:
        name = row.pop(0)
        dist = hash_ = coord = None
        if with_dist:
            dist = float(row.pop(0))
        if with_hash:
            hash_ = int(row.pop(0))
        if with_coord:
            coord = GeoPoint(*map(float, row.pop(0)))

        res_rows.append(GeoMember(name, dist, hash_, coord))

    return res_rows



import types
import asyncio
import socket
from functools import partial
from collections import deque

from .util import (
    encode_command,
    wait_ok,
    _NOTSET,
    _set_result,
    _set_exception,
    coerced_keys_dict,
    decode,
    async_task,
    create_future,
    )
from .parser import Reader
from .errors import (
    ConnectionClosedError,
    RedisError,
    ProtocolError,
    ReplyError,
    WatchVariableError,
    ReadOnlyError,
    )
from .pubsub import Channel
from .abc import AbcChannel
from .abc import AbcConnection
from .log import logger


__all__ = ['create_connection', 'RedisConnection']

MAX_CHUNK_SIZE = 65536

_PUBSUB_COMMANDS = (
    'SUBSCRIBE', b'SUBSCRIBE',
    'PSUBSCRIBE', b'PSUBSCRIBE',
    'UNSUBSCRIBE', b'UNSUBSCRIBE',
    'PUNSUBSCRIBE', b'PUNSUBSCRIBE',
    )


@asyncio.coroutine
def create_connection(address, *, db=None, password=None, ssl=None,
                      encoding=None, parser=None, loop=None, timeout=None,
                      connection_cls=None):
    """Creates redis connection.

    Opens connection to Redis server specified by address argument.
    Address argument is similar to socket address argument, ie:
    * when address is a tuple it represents (host, port) pair;
    * when address is a str it represents unix domain socket path.
    (no other address formats supported)

    SSL argument is passed through to asyncio.create_connection.
    By default SSL/TLS is not used.

    By default any timeout is applied at the connection stage, however
    you can set a limitted time used trying to open a connection via
    the `timeout` Kw.

    Encoding argument can be used to decode byte-replies to strings.
    By default no decoding is done.

    Parser parameter can be used to pass custom Redis protocol parser class.
    By default hiredis.Reader is used (unless it is missing or platform
    is not CPython).

    Return value is RedisConnection instance or a connection_cls if it is
    given.

    This function is a coroutine.
    """
    assert isinstance(address, (tuple, list, str)), "tuple or str expected"

    if timeout is not None and timeout <= 0:
        raise ValueError("Timeout has to be None or a number greater than 0")

    if connection_cls:
        assert issubclass(connection_cls, AbcConnection),\
                "connection_class does not meet the AbcConnection contract"
        cls = connection_cls
    else:
        cls = RedisConnection

    if isinstance(address, (list, tuple)):
        host, port = address
        logger.debug("Creating tcp connection to %r", address)
        reader, writer = yield from asyncio.wait_for(asyncio.open_connection(
            host, port, ssl=ssl, loop=loop), timeout, loop=loop)
        sock = writer.transport.get_extra_info('socket')
        if sock is not None:
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            address = sock.getpeername()
        address = tuple(address[:2])
    else:
        logger.debug("Creating unix connection to %r", address)
        reader, writer = yield from asyncio.wait_for(
            asyncio.open_unix_connection(address, ssl=ssl, loop=loop),
            timeout, loop=loop)
        sock = writer.transport.get_extra_info('socket')
        if sock is not None:
            address = sock.getpeername()

    conn = cls(reader, writer, encoding=encoding,
               address=address, parser=parser,
               loop=loop)

    try:
        if password is not None:
            yield from conn.auth(password)
        if db is not None:
            yield from conn.select(db)
    except Exception:
        conn.close()
        yield from conn.wait_closed()
        raise
    return conn


class RedisConnection(AbcConnection):
    """Redis connection."""

    def __init__(self, reader, writer, *, address, encoding=None,
                 parser=None, loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        if parser is None:
            parser = Reader
        assert callable(parser), (
            "Parser argument is not callable", parser)
        self._reader = reader
        self._writer = writer
        self._address = address
        self._loop = loop
        self._waiters = deque()
        self._parser = parser(protocolError=ProtocolError,
                              replyError=ReplyError)
        self._reader_task = async_task(self._read_data(), loop=self._loop)
        self._db = 0
        self._closing = False
        self._closed = False
        self._close_waiter = create_future(loop=self._loop)
        self._reader_task.add_done_callback(self._close_waiter.set_result)
        self._in_transaction = None
        self._transaction_error = None  # XXX: never used?
        self._in_pubsub = 0
        self._pubsub_channels = coerced_keys_dict()
        self._pubsub_patterns = coerced_keys_dict()
        self._encoding = encoding

    def __repr__(self):
        return '<RedisConnection [db:{}]>'.format(self._db)

    @asyncio.coroutine
    def _read_data(self):
        """Response reader task."""
        while not self._reader.at_eof():
            try:
                data = yield from self._reader.read(MAX_CHUNK_SIZE)
            except asyncio.CancelledError:
                break
            except Exception as exc:
                # XXX: for QUIT command connection error can be received
                #       before response
                logger.error("Exception on data read %r", exc, exc_info=True)
                break
            if data == b'' and self._reader.at_eof():
                logger.debug("Connection has been closed by server")
                break
            self._parser.feed(data)
            while True:
                try:
                    obj = self._parser.gets()
                except ProtocolError as exc:
                    # ProtocolError is fatal
                    # so connection must be closed
                    if self._in_transaction is not None:
                        self._transaction_error = exc
                    self._closing = True
                    self._do_close(exc)
                    return
                else:
                    if obj is False:
                        break
                    if self._in_pubsub:
                        self._process_pubsub(obj)
                    else:
                        self._process_data(obj)
        self._closing = True
        self._do_close(None)

    def _process_data(self, obj):
        """Processes command results."""
        assert len(self._waiters) > 0, (type(obj), obj)
        waiter, encoding, cb = self._waiters.popleft()
        if isinstance(obj, RedisError):
            if isinstance(obj, ReplyError):
                if obj.args[0].startswith('READONLY'):
                    obj = ReadOnlyError(obj.args[0])
            _set_exception(waiter, obj)
            if self._in_transaction is not None:
                self._transaction_error = obj
        else:
            if encoding is not None:
                try:
                    obj = decode(obj, encoding)
                except Exception as exc:
                    _set_exception(waiter, exc)
                    return
            if cb is not None:
                try:
                    obj = cb(obj)
                except Exception as exc:
                    _set_exception(waiter, exc)
                    return
            _set_result(waiter, obj)
            if self._in_transaction is not None:
                self._in_transaction.append((encoding, cb))

    def _process_pubsub(self, obj, *, process_waiters=True):
        """Processes pubsub messages."""
        kind, *pattern, chan, data = obj
        if kind in (b'subscribe', b'unsubscribe'):
            if process_waiters and self._in_pubsub and self._waiters:
                self._process_data(obj)
            if kind == b'unsubscribe':
                ch = self._pubsub_channels.pop(chan, None)
                if ch:
                    ch.close()
            self._in_pubsub = data
        elif kind in (b'psubscribe', b'punsubscribe'):
            if process_waiters and self._in_pubsub and self._waiters:
                self._process_data(obj)
            if kind == b'punsubscribe':
                ch = self._pubsub_patterns.pop(chan, None)
                if ch:
                    ch.close()
            self._in_pubsub = data
        elif kind == b'message':
            self._pubsub_channels[chan].put_nowait(data)
        elif kind == b'pmessage':
            pattern = pattern[0]
            self._pubsub_patterns[pattern].put_nowait((chan, data))
        else:
            logger.warning("Unknown pubsub message received %r", obj)

    def execute(self, command, *args, encoding=_NOTSET):
        """Executes redis command and returns Future waiting for the answer.

        Raises:
        * TypeError if any of args can not be encoded as bytes.
        * ReplyError on redis '-ERR' resonses.
        * ProtocolError when response can not be decoded meaning connection
          is broken.
        """
        if self._reader is None or self._reader.at_eof():
            raise ConnectionClosedError("Connection closed or corrupted")
        if command is None:
            raise TypeError("command must not be None")
        if None in set(args):
            raise TypeError("args must not contain None")
        command = command.upper().strip()
        is_pubsub = command in _PUBSUB_COMMANDS
        if self._in_pubsub and not is_pubsub:
            raise RedisError("Connection in SUBSCRIBE mode")
        elif is_pubsub:
            logger.warning("Deprecated. Use `execute_pubsub` method directly")
            return self.execute_pubsub(command, *args)

        if command in ('SELECT', b'SELECT'):
            cb = partial(self._set_db, args=args)
        elif command in ('MULTI', b'MULTI'):
            cb = self._start_transaction
        elif command in ('EXEC', b'EXEC'):
            cb = partial(self._end_transaction, discard=False)
        elif command in ('DISCARD', b'DISCARD'):
            cb = partial(self._end_transaction, discard=True)
        else:
            cb = None
        if encoding is _NOTSET:
            encoding = self._encoding
        fut = create_future(loop=self._loop)
        self._writer.write(encode_command(command, *args))
        self._waiters.append((fut, encoding, cb))
        return fut

    def execute_pubsub(self, command, *channels):
        """Executes redis (p)subscribe/(p)unsubscribe commands.

        Returns asyncio.gather coroutine waiting for all channels/patterns
        to receive answers.
        """
        command = command.upper().strip()
        assert command in _PUBSUB_COMMANDS, (
            "Pub/Sub command expected", command)
        if self._reader is None or self._reader.at_eof():
            raise ConnectionClosedError("Connection closed or corrupted")
        if None in set(channels):
            raise TypeError("args must not contain None")
        if not len(channels):
            raise TypeError("No channels/patterns supplied")
        is_pattern = len(command) in (10, 12)
        mkchannel = partial(Channel, is_pattern=is_pattern, loop=self._loop)
        channels = [ch if isinstance(ch, AbcChannel) else mkchannel(ch)
                    for ch in channels]
        if not all(ch.is_pattern == is_pattern for ch in channels):
            raise ValueError("Not all channels {} match command {}"
                             .format(channels, command))
        cmd = encode_command(command, *(ch.name for ch in channels))
        res = []
        for ch in channels:
            fut = create_future(loop=self._loop)
            res.append(fut)
            cb = partial(self._update_pubsub, ch=ch)
            self._waiters.append((fut, None, cb))
        self._writer.write(cmd)
        return asyncio.gather(*res, loop=self._loop)

    def close(self):
        """Close connection."""
        self._do_close(None)

    def _do_close(self, exc):
        if self._closed:
            return
        self._closed = True
        self._closing = False
        self._writer.transport.close()
        self._reader_task.cancel()
        self._reader_task = None
        self._writer = None
        self._reader = None
        while self._waiters:
            waiter, *spam = self._waiters.popleft()
            logger.debug("Cancelling waiter %r", (waiter, spam))
            if exc is None:
                waiter.cancel()
            else:
                waiter.set_exception(exc)
        while self._pubsub_channels:
            _, ch = self._pubsub_channels.popitem()
            logger.debug("Closing pubsub channel %r", ch)
            ch.close()
        while self._pubsub_patterns:
            _, ch = self._pubsub_patterns.popitem()
            logger.debug("Closing pubsub pattern %r", ch)
            ch.close()

    @property
    def closed(self):
        """True if connection is closed."""
        closed = self._closing or self._closed
        if not closed and self._reader and self._reader.at_eof():
            self._closing = closed = True
            self._loop.call_soon(self._do_close, None)
        return closed

    @asyncio.coroutine
    def wait_closed(self):
        """Coroutine waiting until connection is closed."""
        yield from asyncio.shield(self._close_waiter, loop=self._loop)

    @property
    def db(self):
        """Currently selected db index."""
        return self._db

    @property
    def encoding(self):
        """Current set codec or None."""
        return self._encoding

    @property
    def address(self):
        """Redis server address, either host-port tuple or str."""
        return self._address

    def select(self, db):
        """Change the selected database for the current connection."""
        if not isinstance(db, int):
            raise TypeError("DB must be of int type, not {!r}".format(db))
        if db < 0:
            raise ValueError("DB must be greater or equal 0, got {!r}"
                             .format(db))
        fut = self.execute('SELECT', db)
        return wait_ok(fut)

    def _set_db(self, ok, args):
        assert ok in {b'OK', 'OK'}, ("Unexpected result of SELECT", ok)
        self._db = args[0]
        return ok

    def _start_transaction(self, ok):
        assert self._in_transaction is None, (
            "Connection is already in transaction", self._in_transaction)
        self._in_transaction = deque()
        self._transaction_error = None
        return ok

    def _end_transaction(self, obj, discard):
        assert self._in_transaction is not None, (
            "Connection is not in transaction", obj)
        self._transaction_error = None
        recall, self._in_transaction = self._in_transaction, None
        recall.popleft()  # ignore first (its _start_transaction)
        if discard:
            return obj
        assert isinstance(obj, list) or (obj is None and not discard), (
            "Unexpected MULTI/EXEC result", obj, recall)
        # TODO: need to be able to re-try transaction
        if obj is None:
            err = WatchVariableError("WATCH variable has changed")
            obj = [err] * len(recall)
        assert len(obj) == len(recall), (
            "Wrong number of result items in mutli-exec", obj, recall)
        res = []
        for o, (encoding, cb) in zip(obj, recall):
            if not isinstance(o, RedisError):
                try:
                    if encoding:
                        o = decode(o, encoding)
                    if cb:
                        o = cb(o)
                except Exception as err:
                    res.append(err)
                    continue
            res.append(o)
        return res

    def _update_pubsub(self, obj, *, ch):
        kind, *pattern, channel, subscriptions = obj
        self._in_pubsub, was_in_pubsub = subscriptions, self._in_pubsub
        if kind == b'subscribe' and channel not in self._pubsub_channels:
            self._pubsub_channels[channel] = ch
        elif kind == b'psubscribe' and channel not in self._pubsub_patterns:
            self._pubsub_patterns[channel] = ch
        if not was_in_pubsub:
            self._process_pubsub(obj, process_waiters=False)
        return obj

    @property
    def in_transaction(self):
        """Set to True when MULTI command was issued."""
        return self._in_transaction is not None

    @property
    def in_pubsub(self):
        """Indicates that connection is in PUB/SUB mode.

        Provides the number of subscribed channels.
        """
        return self._in_pubsub

    @property
    def pubsub_channels(self):
        """Returns read-only channels dict."""
        return types.MappingProxyType(self._pubsub_channels)

    @property
    def pubsub_patterns(self):
        """Returns read-only patterns dict."""
        return types.MappingProxyType(self._pubsub_patterns)

    def auth(self, password):
        """Authenticate to server."""
        fut = self.execute('AUTH', password)
        return wait_ok(fut)



import asyncio
import aioredis


def main():
    loop = asyncio.get_event_loop()

    @asyncio.coroutine
    def go():
        redis = yield from aioredis.create_redis(
            ('localhost', 6379))
        yield from redis.delete('foo', 'bar')
        tr = redis.multi_exec()
        fut1 = tr.incr('foo')
        fut2 = tr.incr('bar')
        res = yield from tr.execute()
        res2 = yield from asyncio.gather(fut1, fut2)
        print(res)
        assert res == res2
        redis.close()
        yield from redis.wait_closed()

    loop.run_until_complete(go())


if __name__ == '__main__':
    main()



import asyncio
import sys

from asyncio.base_events import BaseEventLoop

from .log import logger


PY_35 = sys.version_info >= (3, 5)

_NOTSET = object()


def correct_aiter(func):
    if sys.version_info >= (3, 5, 2):
        return func
    else:
        return asyncio.coroutine(func)


# NOTE: never put here anything else;
#       just this basic types
_converters = {
    bytes: lambda val: val,
    bytearray: lambda val: val,
    str: lambda val: val.encode('utf-8'),
    int: lambda val: str(val).encode('utf-8'),
    float: lambda val: str(val).encode('utf-8'),
    }


def _bytes_len(sized):
    return str(len(sized)).encode('utf-8')


def encode_command(*args):
    """Encodes arguments into redis bulk-strings array.

    Raises TypeError if any of args not of bytes, str, int or float type.
    """
    buf = bytearray()

    def add(data):
        return buf.extend(data + b'\r\n')

    add(b'*' + _bytes_len(args))
    for arg in args:
        if type(arg) in _converters:
            barg = _converters[type(arg)](arg)
            add(b'$' + _bytes_len(barg))
            add(barg)
        else:
            raise TypeError("Argument {!r} expected to be of bytes,"
                            " str, int or float type".format(arg))
    return buf


def decode(obj, encoding):
    if isinstance(obj, bytes):
        return obj.decode(encoding)
    elif isinstance(obj, list):
        return [decode(o, encoding) for o in obj]
    return obj


@asyncio.coroutine
def wait_ok(fut):
    res = yield from fut
    if res in (b'QUEUED', 'QUEUED'):
        return res
    return res in (b'OK', 'OK')


@asyncio.coroutine
def wait_convert(fut, type_, **kwargs):
    result = yield from fut
    if result in (b'QUEUED', 'QUEUED'):
        return result
    return type_(result, **kwargs)


@asyncio.coroutine
def wait_make_dict(fut):
    res = yield from fut
    if res in (b'QUEUED', 'QUEUED'):
        return res
    it = iter(res)
    return dict(zip(it, it))


class coerced_keys_dict(dict):

    def __getitem__(self, other):
        if not isinstance(other, bytes):
            other = _converters[type(other)](other)
        return dict.__getitem__(self, other)

    def __contains__(self, other):
        if not isinstance(other, bytes):
            other = _converters[type(other)](other)
        return dict.__contains__(self, other)


if PY_35:
    class _BaseScanIter:
        __slots__ = ('_scan', '_cur', '_ret')

        def __init__(self, scan):
            self._scan = scan
            self._cur = b'0'
            self._ret = []

        @correct_aiter
        def __aiter__(self):
            return self

    class _ScanIter(_BaseScanIter):

        @asyncio.coroutine
        def __anext__(self):
            while not self._ret and self._cur:
                self._cur, self._ret = yield from self._scan(self._cur)
            if not self._cur and not self._ret:
                raise StopAsyncIteration  # noqa
            else:
                ret = self._ret.pop(0)
                return ret

    class _ScanIterPairs(_BaseScanIter):

        @asyncio.coroutine
        def __anext__(self):
            while not self._ret and self._cur:
                self._cur, ret = yield from self._scan(self._cur)
                self._ret = list(zip(ret[::2], ret[1::2]))
            if not self._cur and not self._ret:
                raise StopAsyncIteration  # noqa
            else:
                ret = self._ret.pop(0)
                return ret


def _set_result(fut, result, *info):
    if fut.done():
        logger.debug("Waiter future is already done %r %r", fut, info)
        assert fut.cancelled(), (
            "waiting future is in wrong state", fut, result, info)
    else:
        fut.set_result(result)


def _set_exception(fut, exception):
    if fut.done():
        logger.debug("Waiter future is already done %r", fut)
        assert fut.cancelled(), (
            "waiting future is in wrong state", fut, exception)
    else:
        fut.set_exception(exception)


if hasattr(asyncio, 'ensure_future'):
    async_task = asyncio.ensure_future
else:
    async_task = asyncio.async  # Deprecated since 3.4.4


# create_future is new in version 3.5.2
if hasattr(BaseEventLoop, 'create_future'):
    def create_future(loop):
        return loop.create_future()
else:
    def create_future(loop):
        return asyncio.Future(loop=loop)



import asyncio
import aioredis


def main():
    loop = asyncio.get_event_loop()

    @asyncio.coroutine
    def go():
        conn = yield from aioredis.create_connection(
            ('localhost', 6379), encoding='utf-8')

        ok = yield from conn.execute('set', 'my-key', 'some value')
        assert ok == 'OK', ok

        str_value = yield from conn.execute('get', 'my-key')
        raw_value = yield from conn.execute('get', 'my-key', encoding=None)
        assert str_value == 'some value'
        assert raw_value == b'some value'

        print('str value:', str_value)
        print('raw value:', raw_value)

        # optionally close connection
        conn.close()
    loop.run_until_complete(go())


if __name__ == '__main__':
    main()



import pytest


@pytest.mark.run_loop
def test_ssl_connection(create_connection, loop, server, ssl_proxy):
    ssl_port, ssl_ctx = ssl_proxy(server.tcp_address.port)

    conn = yield from create_connection(
        ('localhost', ssl_port), ssl=ssl_ctx, loop=loop)
    res = yield from conn.execute('ping')
    assert res == b'PONG'


@pytest.mark.run_loop
def test_ssl_redis(create_redis, loop, server, ssl_proxy):
    ssl_port, ssl_ctx = ssl_proxy(server.tcp_address.port)

    redis = yield from create_redis(
        ('localhost', ssl_port), ssl=ssl_ctx, loop=loop)
    res = yield from redis.ping()
    assert res == b'PONG'


@pytest.mark.run_loop
def test_ssl_pool(create_pool, server, loop, ssl_proxy):
    ssl_port, ssl_ctx = ssl_proxy(server.tcp_address.port)

    pool = yield from create_pool(
        ('localhost', ssl_port), ssl=ssl_ctx, loop=loop)
    with (yield from pool) as conn:
        res = yield from conn.execute('PING')
        assert res == b'PONG'



import pytest

from aioredis import GeoPoint, GeoMember


@pytest.mark.run_loop
@pytest.redis_version(
    3, 2, 0, reason='GEOADD is available since redis >= 3.2.0')
def test_geoadd(redis):
    res = yield from redis.geoadd('geodata', 13.361389, 38.115556, 'Palermo')
    assert res == 1

    res = yield from redis.geoadd(
        'geodata',
        15.087269, 37.502669, 'Catania',
        12.424315, 37.802105, 'Marsala'
    )
    assert res == 2


@pytest.mark.run_loop
@pytest.redis_version(
    3, 2, 0, reason='GEODIST is available since redis >= 3.2.0')
def test_geodist(redis):
    res = yield from redis.geoadd(
        'geodata',
        13.361389, 38.115556, 'Palermo',
        15.087269, 37.502669, 'Catania'
    )
    assert res == 2

    res = yield from redis.geodist('geodata', 'Palermo', 'Catania')
    assert res == 166274.1516

    res = yield from redis.geodist('geodata', 'Palermo', 'Catania', 'km')
    assert res == 166.2742


@pytest.mark.run_loop
@pytest.redis_version(
    3, 2, 0, reason='GEOHASH is available since redis >= 3.2.0')
def test_geohash(redis):
    res = yield from redis.geoadd(
        'geodata',
        13.361389, 38.115556, 'Palermo',
        15.087269, 37.502669, 'Catania'
    )
    assert res == 2

    res = yield from redis.geohash(
        'geodata', 'Palermo', encoding='utf-8'
    )
    assert res == ['sqc8b49rny0']

    res = yield from redis.geohash(
        'geodata', 'Palermo', 'Catania', encoding='utf-8'
    )
    assert res == ['sqc8b49rny0', 'sqdtr74hyu0']


@pytest.mark.run_loop
@pytest.redis_version(
    3, 2, 0, reason='GEOPOS is available since redis >= 3.2.0')
def test_geopos(redis):
    res = yield from redis.geoadd(
        'geodata',
        13.361389, 38.115556, 'Palermo',
        15.087269, 37.502669, 'Catania'
    )
    assert res == 2

    res = yield from redis.geopos('geodata', 'Palermo')
    assert res == [
        GeoPoint(longitude=13.36138933897018433, latitude=38.11555639549629859)
    ]

    res = yield from redis.geopos('geodata', 'Catania', 'Palermo')
    assert res == [
        GeoPoint(longitude=15.087267458438873, latitude=37.50266842333162),
        GeoPoint(longitude=13.36138933897018433, latitude=38.11555639549629859)
    ]


@pytest.mark.run_loop
@pytest.redis_version(
    3, 2, 0, reason='GEO* is available since redis >= 3.2.0')
def test_geo_not_exist_members(redis):
    res = yield from redis.geoadd('geodata', 13.361389, 38.115556, 'Palermo')
    assert res == 1

    res = yield from redis.geoadd(
        'geodata',
        15.087269, 37.502669, 'Catania',
        12.424315, 37.802105, 'Marsala'
    )
    assert res == 2

    res = yield from redis.geohash(
        'geodata', 'NotExistMember'
    )
    assert res == [None]

    res = yield from redis.geodist('geodata', 'NotExistMember', 'Catania')
    assert res is None

    res = yield from redis.geopos(
        'geodata', 'Palermo', 'NotExistMember', 'Catania'
    )
    assert res == [
        GeoPoint(
            longitude=13.36138933897018433,
            latitude=38.11555639549629859
        ),
        None,
        GeoPoint(longitude=15.087267458438873, latitude=37.50266842333162)
    ]


@pytest.mark.run_loop
@pytest.redis_version(
    3, 2, 0, reason='GEORADIUS is available since redis >= 3.2.0')
def test_georadius_validation(redis):
    res = yield from redis.geoadd(
        'geodata',
        13.361389, 38.115556, 'Palermo', 15.087269, 37.502669, 'Catania'
    )
    assert res == 2

    with pytest.raises(TypeError):
        res = yield from redis.georadius(
            'geodata', 15, 37, 200, 'km', count=1.3, encoding='utf-8'
        )
    with pytest.raises(TypeError):
        res = yield from redis.georadius(
            'geodata', 15, 37, '200', 'km', encoding='utf-8'
        )
    with pytest.raises(ValueError):
        res = yield from redis.georadius(
            'geodata', 15, 37, 200, 'k', encoding='utf-8'
        )
    with pytest.raises(ValueError):
        res = yield from redis.georadius(
            'geodata', 15, 37, 200, 'km', sort='DESV', encoding='utf-8'
        )


@pytest.mark.run_loop
@pytest.redis_version(
    3, 2, 0, reason='GEORADIUS is available since redis >= 3.2.0')
def test_georadius(redis):
    res = yield from redis.geoadd(
        'geodata',
        13.361389, 38.115556, 'Palermo', 15.087269, 37.502669, 'Catania'
    )
    assert res == 2

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km', encoding='utf-8'
    )
    assert res == ['Palermo', 'Catania']

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km', count=1, encoding='utf-8'
    )
    assert res == ['Catania']

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km', sort='ASC', encoding='utf-8'
    )
    assert res == ['Catania', 'Palermo']

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km', with_dist=True, encoding='utf-8'
    )
    assert res == [
        GeoMember(member='Palermo', dist=190.4424, coord=None, hash=None),
        GeoMember(member='Catania', dist=56.4413, coord=None, hash=None)
    ]

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km',
        with_dist=True, with_coord=True, encoding='utf-8'
    )
    assert res == [
        GeoMember(
            member='Palermo', dist=190.4424, hash=None,
            coord=GeoPoint(
                longitude=13.36138933897018433, latitude=38.11555639549629859
            )
        ),
        GeoMember(
            member='Catania', dist=56.4413, hash=None,
            coord=GeoPoint(
                longitude=15.087267458438873, latitude=37.50266842333162
            ),
        )
    ]

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km',
        with_dist=True, with_coord=True, with_hash=True, encoding='utf-8'
    )
    assert res == [
        GeoMember(
            member='Palermo', dist=190.4424, hash=3479099956230698,
            coord=GeoPoint(
                longitude=13.36138933897018433, latitude=38.11555639549629859
            )
        ),
        GeoMember(
            member='Catania', dist=56.4413, hash=3479447370796909,
            coord=GeoPoint(
                longitude=15.087267458438873, latitude=37.50266842333162
            ),
        )
    ]

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km',
        with_coord=True, with_hash=True, encoding='utf-8'
    )
    assert res == [
        GeoMember(
            member='Palermo', dist=None, hash=3479099956230698,
            coord=GeoPoint(
                longitude=13.36138933897018433, latitude=38.11555639549629859
            )
        ),
        GeoMember(
            member='Catania', dist=None, hash=3479447370796909,
            coord=GeoPoint(
                longitude=15.087267458438873, latitude=37.50266842333162
            ),
        )
    ]

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km', with_coord=True, encoding='utf-8'
    )
    assert res == [
        GeoMember(
            member='Palermo', dist=None, hash=None,
            coord=GeoPoint(
                longitude=13.36138933897018433, latitude=38.11555639549629859
            )
        ),
        GeoMember(
            member='Catania', dist=None, hash=None,
            coord=GeoPoint(
                longitude=15.087267458438873, latitude=37.50266842333162
            ),
        )
    ]

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km', count=1, sort='DESC',
        with_hash=True, encoding='utf-8'
    )
    assert res == [
        GeoMember(
            member='Palermo', dist=None, hash=3479099956230698, coord=None
        )
    ]


@pytest.mark.run_loop
@pytest.redis_version(
    3, 2, 0, reason='GEORADIUSBYMEMBER is available since redis >= 3.2.0')
def test_georadiusbymember(redis):
    res = yield from redis.geoadd(
        'geodata',
        13.361389, 38.115556, 'Palermo',
        15.087269, 37.502669, 'Catania'
    )
    assert res == 2

    res = yield from redis.georadiusbymember(
        'geodata', 'Palermo', 200, 'km', with_dist=True, encoding='utf-8'
    )
    assert res == [
        GeoMember(member='Palermo', dist=0.0, coord=None, hash=None),
        GeoMember(member='Catania', dist=166.2742, coord=None, hash=None)
    ]
    res = yield from redis.georadiusbymember(
        'geodata', 'Palermo', 200, 'km', encoding='utf-8'
    )
    assert res == ['Palermo', 'Catania']

    res = yield from redis.georadiusbymember(
        'geodata', 'Palermo', 200, 'km',
        with_dist=True, with_coord=True, encoding='utf-8'
    )
    assert res == [
        GeoMember(
            member='Palermo', dist=0.0, hash=None,
            coord=GeoPoint(13.361389338970184, 38.1155563954963)
        ),
        GeoMember(
            member='Catania', dist=166.2742, hash=None,
            coord=GeoPoint(15.087267458438873, 37.50266842333162)
        )
    ]

    res = yield from redis.georadiusbymember(
        'geodata', 'Palermo', 200, 'km',
        with_dist=True, with_coord=True, with_hash=True, encoding='utf-8'
    )
    assert res == [
        GeoMember(
            member='Palermo', dist=0.0, hash=3479099956230698,
            coord=GeoPoint(13.361389338970184, 38.1155563954963)
        ),
        GeoMember(
            member='Catania', dist=166.2742, hash=3479447370796909,
            coord=GeoPoint(15.087267458438873, 37.50266842333162)
        )
    ]


@pytest.mark.run_loop
@pytest.redis_version(
    3, 2, 0, reason='GEOHASH is available since redis >= 3.2.0')
def test_geohash_binary(redis):
    res = yield from redis.geoadd(
        'geodata',
        13.361389, 38.115556, 'Palermo',
        15.087269, 37.502669, 'Catania'
    )
    assert res == 2

    res = yield from redis.geohash(
        'geodata', 'Palermo'
    )
    assert res == [b'sqc8b49rny0']

    res = yield from redis.geohash(
        'geodata', 'Palermo', 'Catania'
    )
    assert res == [b'sqc8b49rny0', b'sqdtr74hyu0']


@pytest.mark.run_loop
@pytest.redis_version(
    3, 2, 0, reason='GEORADIUS is available since redis >= 3.2.0')
def test_georadius_binary(redis):
    res = yield from redis.geoadd(
        'geodata',
        13.361389, 38.115556, 'Palermo', 15.087269, 37.502669, 'Catania'
    )
    assert res == 2

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km'
    )
    assert res == [b'Palermo', b'Catania']

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km', count=1
    )
    assert res == [b'Catania']

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km', sort='ASC'
    )
    assert res == [b'Catania', b'Palermo']

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km', with_dist=True
    )
    assert res == [
        GeoMember(member=b'Palermo', dist=190.4424, coord=None, hash=None),
        GeoMember(member=b'Catania', dist=56.4413, coord=None, hash=None)
    ]

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km',
        with_dist=True, with_coord=True
    )
    assert res == [
        GeoMember(
            member=b'Palermo', dist=190.4424, hash=None,
            coord=GeoPoint(
                longitude=13.36138933897018433, latitude=38.11555639549629859
            )
        ),
        GeoMember(
            member=b'Catania', dist=56.4413, hash=None,
            coord=GeoPoint(
                longitude=15.087267458438873, latitude=37.50266842333162
            ),
        )
    ]

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km',
        with_dist=True, with_coord=True, with_hash=True
    )
    assert res == [
        GeoMember(
            member=b'Palermo', dist=190.4424, hash=3479099956230698,
            coord=GeoPoint(
                longitude=13.36138933897018433, latitude=38.11555639549629859
            )
        ),
        GeoMember(
            member=b'Catania', dist=56.4413, hash=3479447370796909,
            coord=GeoPoint(
                longitude=15.087267458438873, latitude=37.50266842333162
            ),
        )
    ]

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km',
        with_coord=True, with_hash=True
    )
    assert res == [
        GeoMember(
            member=b'Palermo', dist=None, hash=3479099956230698,
            coord=GeoPoint(
                longitude=13.36138933897018433, latitude=38.11555639549629859
            )
        ),
        GeoMember(
            member=b'Catania', dist=None, hash=3479447370796909,
            coord=GeoPoint(
                longitude=15.087267458438873, latitude=37.50266842333162
            ),
        )
    ]

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km', with_coord=True
    )
    assert res == [
        GeoMember(
            member=b'Palermo', dist=None, hash=None,
            coord=GeoPoint(
                longitude=13.36138933897018433, latitude=38.11555639549629859
            )
        ),
        GeoMember(
            member=b'Catania', dist=None, hash=None,
            coord=GeoPoint(
                longitude=15.087267458438873, latitude=37.50266842333162
            ),
        )
    ]

    res = yield from redis.georadius(
        'geodata', 15, 37, 200, 'km', count=1, sort='DESC',
        with_hash=True
    )
    assert res == [
        GeoMember(
            member=b'Palermo', dist=None, hash=3479099956230698, coord=None
        )
    ]


@pytest.mark.run_loop
@pytest.redis_version(
    3, 2, 0, reason='GEORADIUSBYMEMBER is available since redis >= 3.2.0')
def test_georadiusbymember_binary(redis):
    res = yield from redis.geoadd(
        'geodata',
        13.361389, 38.115556, 'Palermo',
        15.087269, 37.502669, 'Catania'
    )
    assert res == 2

    res = yield from redis.georadiusbymember(
        'geodata', 'Palermo', 200, 'km', with_dist=True
    )
    assert res == [
        GeoMember(member=b'Palermo', dist=0.0, coord=None, hash=None),
        GeoMember(member=b'Catania', dist=166.2742, coord=None, hash=None)
    ]

    res = yield from redis.georadiusbymember(
        'geodata', 'Palermo', 200, 'km',
        with_dist=True, with_coord=True
    )
    assert res == [
        GeoMember(
            member=b'Palermo', dist=0.0, hash=None,
            coord=GeoPoint(13.361389338970184, 38.1155563954963)
        ),
        GeoMember(
            member=b'Catania', dist=166.2742, hash=None,
            coord=GeoPoint(15.087267458438873, 37.50266842333162)
        )
    ]

    res = yield from redis.georadiusbymember(
        'geodata', 'Palermo', 200, 'km',
        with_dist=True, with_coord=True, with_hash=True
    )
    assert res == [
        GeoMember(
            member=b'Palermo', dist=0.0, hash=3479099956230698,
            coord=GeoPoint(13.361389338970184, 38.1155563954963)
        ),
        GeoMember(
            member=b'Catania', dist=166.2742, hash=3479447370796909,
            coord=GeoPoint(15.087267458438873, 37.50266842333162)
        )
    ]



import asyncio
import aioredis


async def main():
    # Redis client bound to single connection (no auto reconnection).
    redis = await aioredis.create_redis(
        ('localhost', 6379))
    await redis.set('my-key', 'value')
    val = await redis.get('my-key')
    print(val)

    # gracefully closing underlying connection
    redis.close()
    await redis.wait_closed()


async def redis_pool():
    # Redis client bound to pool of connections (auto-reconnecting).
    redis = await aioredis.create_redis_pool(
        ('localhost', 6379))
    await redis.set('my-key', 'value')
    val = await redis.get('my-key')
    print(val)

    # gracefully closing underlying connection
    redis.close()
    await redis.wait_closed()


if __name__ == '__main__':
    asyncio.get_event_loop().run_until_complete(main())
    asyncio.get_event_loop().run_until_complete(redis_pool())



import asyncio
import aioredis


def main():
    loop = asyncio.get_event_loop()

    @asyncio.coroutine
    def go():
        redis = yield from aioredis.create_redis(
            ('localhost', 6379))
        yield from redis.set('my-key', 'value')
        val = yield from redis.get('my-key')
        print(val)

        # optinally closing underlying connection
        redis.close()
    loop.run_until_complete(go())


if __name__ == '__main__':
    main()



from aioredis.util import wait_ok


class ClusterCommandsMixin:
    """Cluster commands mixin.

    For commands details see: http://redis.io/commands#cluster
    """

    def cluster_add_slots(self, slot, *slots):
        """Assign new hash slots to receiving node."""
        slots = (slot,) + slots
        if not all(isinstance(s, int) for s in slots):
            raise TypeError("All parameters must be of type int")
        fut = self.execute(b'CLUSTER', b'ADDSLOTS', *slots)
        return wait_ok(fut)

    def cluster_count_failure_reports(self, node_id):
        """Return the number of failure reports active for a given node."""
        return self.execute(
            b'CLUSTER', b'COUNT-FAILURE-REPORTS', node_id)

    def cluster_count_key_in_slots(self, slot):
        """Return the number of local keys in the specified hash slot."""
        if not isinstance(slot, int):
            raise TypeError("Expected slot to be of type int, got {}"
                            .format(type(slot)))
        return self.execute(b'CLUSTER', b'COUNTKEYSINSLOT', slot)

    def cluster_del_slots(self, slot, *slots):
        """Set hash slots as unbound in receiving node."""
        slots = (slot,) + slots
        if not all(isinstance(s, int) for s in slots):
            raise TypeError("All parameters must be of type int")
        fut = self.execute(b'CLUSTER', b'DELSLOTS', *slots)
        return wait_ok(fut)

    def cluster_failover(self):
        """Forces a slave to perform a manual failover of its master."""
        pass    # TODO: Implement

    def cluster_forget(self, node_id):
        """Remove a node from the nodes table."""
        fut = self.execute(b'CLUSTER', b'FORGET', node_id)
        return wait_ok(fut)

    def cluster_get_keys_in_slots(self, slot, count, *, encoding):
        """Return local key names in the specified hash slot."""
        return self.execute(b'CLUSTER', b'GETKEYSINSLOT', slot, count,
                            encoding=encoding)

    def cluster_info(self):
        """Provides info about Redis Cluster node state."""
        pass    # TODO: Implement

    def cluster_keyslot(self, key):
        """Returns the hash slot of the specified key."""
        return self.execute(b'CLUSTER', b'KEYSLOT', key)

    def cluster_meet(self, ip, port):
        """Force a node cluster to handshake with another node."""
        fut = self.execute(b'CLUSTER', b'MEET', ip, port)
        return wait_ok(fut)

    def cluster_nodes(self):
        """Get Cluster config for the node."""
        pass    # TODO: Implement

    def cluster_replicate(self, node_id):
        """Reconfigure a node as a slave of the specified master node."""
        fut = self.execute(b'CLUSTER', b'REPLICATE', node_id)
        return wait_ok(fut)

    def cluster_reset(self, *, hard=False):
        """Reset a Redis Cluster node."""
        reset = hard and b'HARD' or b'SOFT'
        fut = self.execute(b'CLUSTER', b'RESET', reset)
        return wait_ok(fut)

    def cluster_save_config(self):
        """Force the node to save cluster state on disk."""
        fut = self.execute(b'CLUSTER', b'SAVECONFIG')
        return wait_ok(fut)

    def cluster_set_config_epoch(self, config_epoch):
        """Set the configuration epoch in a new node."""
        fut = self.execute(b'CLUSTER', b'SET-CONFIG-EPOCH', config_epoch)
        return wait_ok(fut)

    def cluster_setslot(self, slot, command, node_id):
        """Bind a hash slot to specified node."""
        pass    # TODO: Implement

    def cluster_slaves(self, node_id):
        """List slave nodes of the specified master node."""
        pass    # TODO: Implement

    def cluster_slots(self):
        """Get array of Cluster slot to node mappings."""
        pass    # TODO: Implement



import pytest

import asyncio


@pytest.mark.run_loop
def test_future_cancellation(create_connection, loop, server):
    conn = yield from create_connection(
        server.tcp_address, loop=loop)

    ts = loop.time()
    fut = conn.execute('BLPOP', 'some-list', 5)
    with pytest.raises(asyncio.TimeoutError):
        yield from asyncio.wait_for(fut, 1, loop=loop)
    assert fut.cancelled()

    # NOTE: Connection becomes available only after timeout expires
    yield from conn.execute('TIME')
    dt = int(loop.time() - ts)
    assert dt in {4, 5, 6}
    # self.assertAlmostEqual(dt, 5.0, delta=1)  # this fails too often



import asyncio
from unittest import mock

from aioredis.commands import MultiExec
from aioredis.commands import Redis
from aioredis.util import create_future


def test_global_loop():
    conn = mock.Mock(spec=(
        'execute closed _transaction_error'
        .split()))
    try:
        old_loop = asyncio.get_event_loop()
    except (AssertionError, RuntimeError):
        old_loop = None
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    tr = MultiExec(conn, commands_factory=Redis)
    assert tr._loop is loop

    def make_fut(cmd, *args, **kw):
        fut = create_future(asyncio.get_event_loop())
        if cmd == 'PING':
            fut.set_result(b'QUEUED')
        elif cmd == 'EXEC':
            fut.set_result([b'PONG'])
        else:
            fut.set_result(b'OK')
        return fut

    conn.execute.side_effect = make_fut
    conn.closed = False
    conn._transaction_error = None

    @asyncio.coroutine
    def go():
        tr.ping()
        res = yield from tr.execute()
        assert res == [b'PONG']
    loop.run_until_complete(go())
    asyncio.set_event_loop(old_loop)
